# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("d3dcsx.dll")
prototypes = \
    {
        # 
        'D3DX11CreateScan': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DX11Scan"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "MaxElementScanSize", "MaxScanCount", "ppScan"]),
        # 
        'D3DX11CreateSegmentedScan': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DX11SegmentedScan"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "MaxElementScanSize", "ppScan"]),
        # 
        'D3DX11CreateFFT': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypePointer(SimStruct({"NumDimensions": SimTypeInt(signed=False, label="UInt32"), "ElementLengths": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 32), "DimensionMask": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="D3DX11_FFT_DATA_TYPE")}, name="D3DX11_FFT_DESC", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "pDesc", "Flags", "pBufferInfo", "ppFFT"]),
        # 
        'D3DX11CreateFFT1DReal': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "X", "Flags", "pBufferInfo", "ppFFT"]),
        # 
        'D3DX11CreateFFT1DComplex': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "X", "Flags", "pBufferInfo", "ppFFT"]),
        # 
        'D3DX11CreateFFT2DReal': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "X", "Y", "Flags", "pBufferInfo", "ppFFT"]),
        # 
        'D3DX11CreateFFT2DComplex': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "X", "Y", "Flags", "pBufferInfo", "ppFFT"]),
        # 
        'D3DX11CreateFFT3DReal': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "X", "Y", "Z", "Flags", "pBufferInfo", "ppFFT"]),
        # 
        'D3DX11CreateFFT3DComplex': SimTypeFunction([SimTypeBottom(label="ID3D11DeviceContext"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"NumTempBufferSizes": SimTypeInt(signed=False, label="UInt32"), "TempBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4), "NumPrecomputeBufferSizes": SimTypeInt(signed=False, label="UInt32"), "PrecomputeBufferFloatSizes": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 4)}, name="D3DX11_FFT_BUFFER_INFO", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="ID3DX11FFT"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pDeviceContext", "X", "Y", "Z", "Flags", "pBufferInfo", "ppFFT"]),
    }

lib.set_prototypes(prototypes)



import angr

######################################
# set_tid_address
######################################

#pylint:disable=redefined-builtin,arguments-differ
class set_tid_address(angr.SimProcedure):
    def run(self, tidptr):
        return 1  # Assume it's single-threaded, so only tid 1 exists



import angr
import select

######################################
# poll
######################################

class poll(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, fds, nfds, timeout):  # pylint: disable=unused-argument
        try:
            nfds_v = self.state.solver.eval_one(nfds)
        except angr.errors.SimSolverError as e:
            raise angr.errors.SimProcedureArgumentError("Can't handle symbolic pollfd arguments") from e
        ###
        # struct pollfd {
        #     int   fd;         /* file descriptor */
        #     short events;     /* requested events */
        #     short revents;    /* returned events */
        # };
        size_of_pollfd = 8
        offset_fd = 0
        offset_events = 4
        offset_revents = 6

        pollfd_array = []
        for offset in range(0, nfds_v):
            pollfd = {
                "fd": self.state.memory.load(fds + offset * size_of_pollfd + offset_fd, 4, endness=self.arch.memory_endness),
                "events": self.state.memory.load(fds + offset * size_of_pollfd + offset_events, 2, endness=self.arch.memory_endness),
                "revents": self.state.memory.load(fds + offset * size_of_pollfd + offset_revents, 2, endness=self.arch.memory_endness)
            }
            pollfd_array.append(pollfd)
        for (offset,pollfd) in enumerate(pollfd_array):
            try:
                fd = self.state.solver.eval_one(pollfd["fd"])
                events = self.state.solver.eval_one(pollfd["events"])
            except angr.errors.SimSolverError as e:
                raise angr.errors.SimProcedureArgumentError("Can't handle symbolic pollfd arguments") from e

            if events & select.POLLIN and fd >= 0:
                revents = pollfd["revents"][self.arch.sizeof["short"]-1:1].concat(self.state.solver.BVS('fd_POLLIN', 1))
                self.state.memory.store(fds + offset * size_of_pollfd + offset_revents, revents, endness=self.arch.memory_endness)

        retval = self.state.solver.BVV(0, 1).concat(self.state.solver.BVS('poll_ret', 31))
        return retval



# pylint: disable=cell-var-from-loop
import concurrent.futures
import logging

from . import ExplorationTechnique
from ..engines.engine import TLSMixin
from ..misc.ux import once

l = logging.getLogger(__name__)

class Threading(ExplorationTechnique):
    """
    Enable multithreading.

    This is only useful in paths where a lot of time is taken inside z3, doing constraint solving.
    This is because of python's GIL, which says that only one thread at a time may be executing python code.
    """
    def __init__(self, threads=8, local_stash='thread_local'):
        super(Threading, self).__init__()
        self.threads = threads
        self.queued = set()
        self.tasks = set()
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=threads)
        self.local_stash = local_stash

    def step(self, simgr, stash='active', error_list=None, target_stash=None, **kwargs):
        target_stash = target_stash or stash
        if error_list is not None:
            raise ValueError("Can't pass error_list to step with threading enabled. Did you install threading twice?")

        l.info("Thread-stepping %s of %s", stash, simgr)

        for state in simgr.stashes[stash]:
            if state in self.queued:
                continue

            # construct new simgr with new lists
            # this means that threads won't trample each other's hook stacks
            # but can still negotiate over shared resources
            tsimgr = simgr.copy()
            tsimgr._stashes = {self.local_stash: [state]}
            tsimgr._errored = []
            self.tasks.add(self.executor.submit(self.inner_step, state, tsimgr, target_stash=target_stash, **kwargs))
            self.queued.add(state)

        timeout = None
        while True:
            done, self.tasks = concurrent.futures.wait(
                self.tasks,
                timeout=timeout,
                return_when=concurrent.futures.FIRST_COMPLETED
            )
            if not done:
                break

            for done_future in done:
                done_future: concurrent.futures.Future
                state, error_list, tsimgr = done_future.result()
                simgr.absorb(tsimgr)
                simgr.errored.extend(error_list)
                simgr.stashes[stash].remove(state)
                self.queued.remove(state)
            timeout = 0

        return simgr

    def inner_step(self, state, simgr, **kwargs):
        error_list = []
        simgr.step(stash=self.local_stash, error_list=error_list, **kwargs)
        return state, error_list, simgr

    def successors(self, simgr, state, engine=None, **kwargs):
        engine = engine or self.project.factory.default_engine
        if not isinstance(engine, TLSMixin) and once("tls_engine"):
            l.error("Using Threading exploration technique but your engine is not thread-safe.")
            l.error("Do you want to add the TLSMixin to your engine?")
        return simgr.successors(state, engine=engine, **kwargs)



# pylint:disable=unused-argument,arguments-differ
import ailment

from ..sequence_walker import SequenceWalker
from ..structurer_nodes import SequenceNode, CodeNode, MultiNode, LoopNode, ConditionNode, EmptyBlockNotice, \
    CascadingConditionNode
from ..condition_processor import ConditionProcessor
from ..utils import insert_node


class IfElseFlattener(SequenceWalker):
    """
    Remove unnecessary else branches and make the else node a direct successor of the previous If node if the If node
    always returns.
    """
    def __init__(self, node, functions):
        handlers = {
            SequenceNode: self._handle_Sequence,
            CodeNode: self._handle_Code,
            MultiNode: self._handle_MultiNode,
            LoopNode: self._handle_Loop,
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
        }

        super().__init__(handlers)
        self.functions = functions
        self.walk(node)

    def _handle_Condition(self, node: ConditionNode, parent=None, index=None, **kwargs):

        if node.true_node is not None and node.false_node is not None:
            try:
                last_stmts = ConditionProcessor.get_last_statements(node.true_node)
            except EmptyBlockNotice:
                last_stmts = None
            if last_stmts is not None \
                    and None not in last_stmts \
                    and all(self._is_statement_terminating(stmt) for stmt in last_stmts):
                # all end points in the true node are returning

                # remove the else node and make it a new node following node
                else_node = node.false_node
                node.false_node = None
                insert_node(parent, index + 1, else_node, index)

                self._handle(else_node, parent=parent, index=index + 1)

        if node.true_node is not None:
            self._handle(node.true_node, parent=node, index=0)
        if node.false_node is not None:
            self._handle(node.false_node, parent=node, index=1)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, parent=None, index=None, **kwargs):
        if node.else_node is not None:
            last_stmts = [ ]
            for _, subnode in node.condition_and_nodes:
                try:
                    last_stmts.extend(ConditionProcessor.get_last_statements(subnode))
                except EmptyBlockNotice:
                    last_stmts.append(None)

            if last_stmts and None not in last_stmts \
                    and all(self._is_statement_terminating(stmt) for stmt in last_stmts):
                # all end points in the true node are returning

                # remove the else node and make it a new node following node
                else_node = node.else_node
                node.else_node = None
                insert_node(parent, index + 1, else_node, index)

                self._handle(else_node, parent=parent, index=index + 1)

        super()._handle_CascadingCondition(node, parent=parent, index=index, **kwargs)

    def _is_statement_terminating(self, stmt):

        if isinstance(stmt, ailment.Stmt.Return):
            return True
        if isinstance(stmt, ailment.Stmt.Call) and isinstance(stmt.target, ailment.Expr.Const):
            # is it calling a non-returning function?
            target_func_addr = stmt.target.value
            try:
                func = self.functions.get_by_addr(target_func_addr)
                return func.returning is False
            except KeyError:
                pass
        return False



from ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class AMulConstDivShrConst(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "(A * N0 / N1) >> N2 => (A * (N0 / 2 ** N2) / N1)"
    expr_classes = (BinaryOp, )

    def optimize(self, expr: BinaryOp):

        if expr.op == "Shr" and len(expr.operands) == 2 and isinstance(expr.operands[1], Const) \
                and isinstance(expr.operands[0], BinaryOp) and expr.operands[0].op == "Div" \
                and isinstance(expr.operands[0].operands[1], Const):
            inner = expr.operands[0].operands[0]
            if isinstance(inner, BinaryOp) and inner.op == "Mul" and isinstance(inner.operands[1], Const):
                a = inner.operands[0]
                N0 = inner.operands[1].value
                N1 = expr.operands[0].operands[1]
                N2 = expr.operands[1].value

                mul = BinaryOp(inner.idx, "Mul",
                               [a,
                                Const(None, None, N0 // (2 ** N2), expr.bits, **expr.operands[0].operands[1].tags)
                                ],
                               False,
                               **inner.tags)
                div = BinaryOp(expr.idx, "Div",
                               [mul,
                                N1],
                               False,
                               **expr.tags)
                return div

        return None



from typing import Optional, Set, List, Tuple, TYPE_CHECKING, Dict
import logging

import networkx

import ailment

from ..codenode import BlockNode
from ..sim_variable import SimMemoryVariable
from ..knowledge_plugins.functions import Function
from .analysis import Analysis, AnalysesHub
from .decompiler.ailblock_walker import AILBlockWalker

if TYPE_CHECKING:
    from angr.analyses.decompiler.decompiler import Decompiler
    from angr.knowledge_plugins.cfg import CFGModel
    from angr.knowledge_plugins.xrefs import XRefManager

_l = logging.getLogger(name=__name__)


class ProxiNodeTypes:
    """
    Node Type Enums
    """
    Empty = 0
    String = 1
    Function = 2
    FunctionCall = 3
    Integer = 4
    Unknown = 5
    Variable = 6


class BaseProxiNode:
    """
    Base class for all nodes in a proximity graph.
    """

    def __init__(self, type_: int, ref_at: Optional[Set[int]] = None):
        self.type_ = type_
        self.ref_at = ref_at

    def __eq__(self, other):
        return isinstance(other, BaseProxiNode) and other.type_ == self.type_ and self.ref_at == other.ref_at

    def __hash__(self):
        return hash((BaseProxiNode, self.type_))


class FunctionProxiNode(BaseProxiNode):
    """
    Proximity node showing current and expanded function calls in graph.
    """

    def __init__(self, func, ref_at: Optional[Set[int]] = None):
        super().__init__(ProxiNodeTypes.Function, ref_at=ref_at)
        self.func = func

    def __eq__(self, other):
        return isinstance(other, FunctionProxiNode) and \
               other.type_ == self.type_ and \
               self.func == other.func

    def __hash__(self):
        return hash((FunctionProxiNode, self.func))


class VariableProxiNode(BaseProxiNode):
    """
    Variable arg node
    """

    def __init__(self, addr, name, ref_at: Optional[Set[int]] = None):
        super().__init__(ProxiNodeTypes.Variable, ref_at=ref_at)
        self.addr = addr
        self.name = name

    def __eq__(self, other):
        return isinstance(other, VariableProxiNode) and \
               other.type_ == self.type_ and \
               self.addr == other.addr

    def __hash__(self):
        return hash((VariableProxiNode, self.addr))


class StringProxiNode(BaseProxiNode):
    """
    String arg node
    """

    def __init__(self, addr, content, ref_at: Optional[Set[int]] = None):
        super().__init__(ProxiNodeTypes.String, ref_at=ref_at)
        self.addr = addr
        self.content = content

    def __eq__(self, other):
        return isinstance(other, StringProxiNode) and \
               other.type_ == self.type_ and \
               self.addr == other.addr

    def __hash__(self):
        return hash((StringProxiNode, self.addr))


class CallProxiNode(BaseProxiNode):
    """
    Call node
    """

    def __init__(self, callee, ref_at: Optional[Set[int]] = None, args: Optional[Tuple[BaseProxiNode]] = None):
        super().__init__(ProxiNodeTypes.FunctionCall, ref_at=ref_at)
        self.callee = callee
        self.args = args

    def __eq__(self, other):
        return isinstance(other, CallProxiNode) and \
               other.type_ == self.type_ and \
               self.callee == other.callee and \
               self.args == other.args and \
               self.ref_at == other.ref_at

    def __hash__(self):
        return hash((CallProxiNode, self.callee, self.args))


class IntegerProxiNode(BaseProxiNode):
    """
    Int arg node
    """

    def __init__(self, value: int, ref_at: Optional[Set[int]] = None):
        super().__init__(ProxiNodeTypes.Integer, ref_at=ref_at)
        self.value = value

    def __eq__(self, other):
        return isinstance(other, IntegerProxiNode) and \
               self.type_ == other.type_ and \
               self.value == other.value

    def __hash__(self):
        return hash((IntegerProxiNode, self.value))


class UnknownProxiNode(BaseProxiNode):
    """
    Unknown arg node
    """

    def __init__(self, dummy_value: str):
        super().__init__(ProxiNodeTypes.Unknown)
        self.dummy_value = dummy_value

    def __eq__(self, other):
        return isinstance(other, UnknownProxiNode) and \
               self.type_ == other.type_ and \
               self.dummy_value == other.dummy_value

    def __hash__(self):
        return hash((UnknownProxiNode, self.dummy_value))


class ProximityGraphAnalysis(Analysis):
    """
    Generate a proximity graph.
    """

    def __init__(self, func: 'Function', cfg_model: 'CFGModel', xrefs: 'XRefManager',
                 decompilation: Optional['Decompiler'] = None, expand_funcs: Optional[Set[int]] = None):
        self._function = func
        self._cfg_model = cfg_model
        self._xrefs = xrefs
        self._decompilation = decompilation
        self._expand_funcs = expand_funcs.copy() if expand_funcs else None

        self.graph: Optional[networkx.DiGraph] = None
        self.handled_stmts = []

        self._work()

    def _work(self):
        self.graph = networkx.DiGraph()

        # initial function
        func_proxi_node = FunctionProxiNode(self._function)

        if not self._decompilation:
            to_expand = self._process_function(self._function, self.graph, func_proxi_node=func_proxi_node)
        else:
            to_expand = self._process_decompilation(self.graph, decompilation=self._decompilation,
                                                    func_proxi_node=func_proxi_node)

        for func_node in to_expand:
            if self._expand_funcs:
                self._expand_funcs.discard(func_node.func.addr)

            subgraph = networkx.DiGraph()
            dec = self._decompilation.project.analyses.Decompiler(func_node.func, cfg=self._decompilation._cfg)
            if not dec:
                sub_expand = self._process_function(func_node.func, subgraph, func_proxi_node=func_node)
            else:
                sub_expand = self._process_decompilation(subgraph, decompilation=dec, func_proxi_node=func_node)

            to_expand.extend(sub_expand)

            # merge subgraph into the original graph
            self.graph.add_nodes_from(subgraph.nodes())
            self.graph.add_edges_from(subgraph.edges())

    def _endnode_connector(self, func: 'Function', subgraph: networkx.DiGraph):
        """
        Properly connect expanded function call's to proximity graph.
        """
        current_node = None
        successors = []
        # Get successor node of the current function node
        for node in self.graph.nodes():
            if isinstance(node, FunctionProxiNode) and node.func == func:
                current_node = node
                successors = list(self.graph.succ[node])
                # Remove edge FunctionProxiNode->successors
                for succ in successors:
                    self.graph.remove_edge(node, succ)
                break

        if successors:
            # add edges subgraph_end_nodes->successor
            end_nodes = [n for n in subgraph.nodes() if subgraph.in_degree(n) >= 1 and subgraph.out_degree(n) == 0]

            # handle subgraphs that are empty
            if not end_nodes:
                end_nodes.append(current_node)
            for end_node in end_nodes:
                for succ in successors:
                    subgraph.add_edge(end_node, succ)

    def _process_function(self, func: 'Function', graph: networkx.DiGraph,
                          func_proxi_node: Optional[FunctionProxiNode] = None) -> List[FunctionProxiNode]:

        to_expand: List[FunctionProxiNode] = []
        found_blocks: Dict[BlockNode: BaseProxiNode] = {}

        # function calls
        for n_ in func.nodes:
            if isinstance(n_, Function):
                func_node = n_
                for block, _, data in func.transition_graph.in_edges(func_node, data=True):
                    if 'ins_addr' in data:
                        if self._expand_funcs and func_node.addr in self._expand_funcs:  # pylint:disable=unsupported-membership-test
                            node = FunctionProxiNode(func_node, ref_at={data['ins_addr']})
                            to_expand.append(node)
                        else:
                            node = CallProxiNode(func_node, ref_at={data['ins_addr']})
                        found_blocks[block] = node

        # subgraph check - do before in case of recursion
        if self.graph == graph:
            subgraph = False
        else:
            subgraph = True

        for edge in func.graph.edges:
            nodes = ()
            for block in edge:
                if block in found_blocks:
                    nodes += (found_blocks[block],)
                else:
                    nodes += (BaseProxiNode(ProxiNodeTypes.Empty, {block.addr}),)
            graph.add_edge(*nodes)

        # Append FunctionProxiNode before Graph
        root_node = [n for n, d in graph.in_degree() if d == 0]
        if root_node:
            graph.add_edge(func_proxi_node, root_node[0])

        # Draw edge from subgraph endnodes to current node's successor
        if subgraph:
            self._endnode_connector(func_proxi_node.func, graph)

        return to_expand

    # TODO look into harnessing ailblock_walker to find Strings, Constants, or Function Calls
    def _arg_handler(self, arg, args, string_refs):
        if isinstance(arg, ailment.Expr.Const):
            # is it a reference to a string?
            if arg.value in self._cfg_model.memory_data:
                md = self._cfg_model.memory_data[arg.value]
                if md.sort == "string":
                    # Yes!
                    args.append(StringProxiNode(arg.value, md.content))
                    string_refs.add(arg.value)
            else:
                # not a string. present it as a constant integer
                args.append(IntegerProxiNode(arg.value, None))
        elif isinstance(arg, ailment.expression.Load):
            if arg.variable is not None:
                args.append(VariableProxiNode(arg.variable.addr, arg.variable.name))
            elif arg.addr.variable is not None:
                if isinstance(arg.addr.variable, SimMemoryVariable):
                    args.append(VariableProxiNode(arg.addr.variable.addr, arg.addr.variable.name))
                else:
                    args.append(VariableProxiNode(None, arg.addr.variable.name))
            else:
                args.append(UnknownProxiNode("l_"))
        elif isinstance(arg, ailment.expression.StackBaseOffset):
            if arg.variable is not None:
                args.append(VariableProxiNode(arg.variable, arg.variable.name))
            else:
                args.append(UnknownProxiNode("s_"))
        else:
            args.append(UnknownProxiNode("_"))

    def _process_decompilation(self, graph: networkx.DiGraph, decompilation: 'Decompiler',
                               func_proxi_node: Optional[FunctionProxiNode] = None) -> List[FunctionProxiNode]:
        to_expand: List[FunctionProxiNode] = []

        # dedup
        string_refs: Set[int] = set()

        # Walk the clinic structure to dump string references and function calls
        ail_graph = decompilation.clinic.cc_graph

        def _handle_Call(stmt_idx: int, stmt: ailment.Stmt.Call,  # pylint:disable=unused-argument
                         block: Optional[ailment.Block]):  # pylint:disable=unused-argument
            func_node = self.kb.functions[stmt.target.value]
            ref_at = {stmt.ins_addr}

            # extract arguments
            args = []
            if stmt.args:
                for arg in stmt.args:
                    self._arg_handler(arg, args, string_refs)

            if self._expand_funcs and func_node.addr in self._expand_funcs:  # pylint:disable=unsupported-membership-test
                new_node = FunctionProxiNode(func_node, ref_at=ref_at)
                if new_node not in to_expand:
                    to_expand.append(new_node)
            else:
                new_node = CallProxiNode(func_node, ref_at=ref_at, args=tuple(args) if args is not None else None)

            # stmt has been properly handled, add the proxi node to the list
            self.handled_stmts.append(new_node)

        # This should have the same functionality as the previous handler
        def _handle_CallExpr(expr_idx: int, expr: ailment.Stmt.Call, stmt_idx: int,  # pylint:disable=unused-argument
                             stmt: ailment.Stmt.Statement,  # pylint:disable=unused-argument
                             block: Optional[ailment.Block]):  # pylint:disable=unused-argument
            _handle_Call(stmt_idx, expr, block)

        # subgraph check - do before in case of recursion
        if self.graph == graph:
            subgraph = False
        else:
            subgraph = True

        # Keep all default handlers, but overwrite necessary ones:
        bw = AILBlockWalker()
        bw.stmt_handlers[ailment.Stmt.Call] = _handle_Call
        bw.expr_handlers[ailment.Stmt.Call] = _handle_CallExpr

        # Custom Graph walker, go through AIL edges
        for ail_edge in ail_graph.edges:
            new_edge = ()
            # Handle both blocks in edge
            for block in ail_edge:
                bw.walk(block)

                # Check to see if our custom _handle's already dealt with the nodes
                if self.handled_stmts:
                    # Add each handled stmt to the graph in a linear fashion []->[]->[]
                    # Linear is fine since stmts come from the same block and are handled in order of appearance
                    for idx, current in enumerate(self.handled_stmts):
                        # Skip Head of List to avoid out of bounds error
                        if idx > 0:
                            graph.add_edge(self.handled_stmts[idx - 1], current)
                    # If first block in edge, LAST handled node -> next node
                    # Else (second block in edge), prev node -> FIRST handled node
                    if block == ail_edge[0]:
                        proxi_node = self.handled_stmts[-1]
                    else:
                        proxi_node = self.handled_stmts[0]

                    # Clear the handled stmts for next block
                    self.handled_stmts = []

                # Block stmts weren't handled, make a simple empty proxi node
                else:
                    proxi_node = BaseProxiNode(ProxiNodeTypes.Empty, {block.addr})

                # Create edge from block_1 (proxi_node_1) -> block_2 (proxi_node_2)
                # both blocks may have multiple proxi nodes, but these are already added to the graph
                new_edge += (proxi_node,)

            # Add the new edge for this pair of blocks
            graph.add_edge(*new_edge)

        # Append FunctionProxiNode before Graph
        root_node = [n for n, d in graph.in_degree() if d == 0]
        if root_node:
            graph.add_edge(func_proxi_node, root_node[0])

        # Draw edge from subgraph endnodes to current node's successor
        if subgraph:
            self._endnode_connector(func_proxi_node.func, graph)

        return to_expand


AnalysesHub.register_default('Proximity', ProximityGraphAnalysis)



# pylint:disable=missing-class-docstring

class AngrError(Exception):
    pass

class AngrValueError(AngrError, ValueError):
    pass

class AngrLifterError(AngrError):
    pass

class AngrExitError(AngrError):
    pass

class AngrPathError(AngrError):
    pass

class AngrVaultError(AngrError):
    pass

class PathUnreachableError(AngrPathError):
    pass

class SimulationManagerError(AngrError):
    pass

class AngrInvalidArgumentError(AngrError):
    pass

class AngrSurveyorError(AngrError):
    pass

class AngrAnalysisError(AngrError):
    pass

class AngrBladeError(AngrError):
    pass

class AngrBladeSimProcError(AngrBladeError):
    pass

class AngrAnnotatedCFGError(AngrError):
    pass

class AngrBackwardSlicingError(AngrError):
    pass

class AngrGirlScoutError(AngrError):
    pass

class AngrCallableError(AngrSurveyorError):
    pass

class AngrCallableMultistateError(AngrCallableError):
    pass

class AngrSyscallError(AngrError):
    pass

class AngrSimOSError(AngrError):
    pass

class AngrAssemblyError(AngrError):
    pass

# Congruency check failure
class AngrIncongruencyError(AngrAnalysisError):
    pass

#
# ForwardAnalysis errors
#

class AngrForwardAnalysisError(AngrError):
    pass

class AngrSkipJobNotice(AngrForwardAnalysisError):
    pass

class AngrDelayJobNotice(AngrForwardAnalysisError):
    pass

class AngrJobMergingFailureNotice(AngrForwardAnalysisError):
    pass

class AngrJobWideningFailureNotice(AngrForwardAnalysisError):
    pass

#
# CFG errors
#

class AngrCFGError(AngrError):
    pass

#
# VFG Errors and notices
#

class AngrVFGError(AngrError):
    pass

class AngrVFGRestartAnalysisNotice(AngrVFGError):
    pass

#
# Data graph errors
#

class AngrDataGraphError(AngrAnalysisError):
    # TODO: deprecated
    pass

class AngrDDGError(AngrAnalysisError):
    pass

#
# Loop analysis
#

class AngrLoopAnalysisError(AngrAnalysisError):
    pass

#
# Exploration techniques
#

class AngrExplorationTechniqueError(AngrError):
    pass

class AngrExplorerError(AngrExplorationTechniqueError):
    pass

class AngrDirectorError(AngrExplorationTechniqueError):
    pass

class AngrTracerError(AngrExplorationTechniqueError):
    pass


#
# VariableRecovery errors
#

class AngrVariableRecoveryError(AngrAnalysisError):
    pass


#
# AngrDB errors
#


class AngrDBError(AngrError):
    pass


class AngrCorruptDBError(AngrDBError):
    pass

class AngrIncompatibleDBError(AngrDBError):
    pass

#
# Tracer
#

class TracerEnvironmentError(AngrError):
    pass

#
# Simulation errors
#

class SimError(Exception):
    bbl_addr = None
    stmt_idx = None
    ins_addr = None
    executed_instruction_count = None
    guard = None

    def record_state(self, state):
        self.bbl_addr = state.scratch.bbl_addr
        self.stmt_idx = state.scratch.stmt_idx
        self.ins_addr = state.scratch.ins_addr
        self.executed_instruction_count = state.history.recent_instruction_count
        self.guard = state.scratch.guard
        return self

#
# State-related errors
#

class SimStateError(SimError):
    pass

class SimMergeError(SimStateError):
    pass

class SimMemoryError(SimStateError):
    pass

class SimMemoryMissingError(SimMemoryError):
    def __init__(self, missing_addr, missing_size, *args):
        super().__init__(missing_addr, missing_size, *args)
        self.missing_addr = missing_addr
        self.missing_size = missing_size

class SimAbstractMemoryError(SimMemoryError):
    pass

class SimRegionMapError(SimMemoryError):
    pass

class SimMemoryLimitError(SimMemoryError):
    pass

class SimMemoryAddressError(SimMemoryError):
    pass

class SimFastMemoryError(SimMemoryError):
    pass

class SimEventError(SimStateError):
    pass

class SimPosixError(SimStateError):
    pass


class SimFilesystemError(SimError):
    pass

class SimSymbolicFilesystemError(SimFilesystemError):
    pass

class SimFileError(SimMemoryError, SimFilesystemError):
    pass

class SimHeapError(SimStateError):
    pass

#
# Error class during VEX parsing
#

class SimUnsupportedError(SimError):
    pass

#
# Solver-related errors
#

class SimSolverError(SimError):
    pass

class SimSolverModeError(SimSolverError):
    pass

class SimSolverOptionError(SimSolverError):
    pass

class SimValueError(SimSolverError):
    pass

class SimUnsatError(SimValueError):
    pass

#
# SimIROp errors
#

class SimOperationError(SimError):
    pass

class UnsupportedIROpError(SimOperationError, SimUnsupportedError):
    pass

#
# SimIRExpr errors
#

class SimExpressionError(SimError):
    pass

class UnsupportedIRExprError(SimExpressionError, SimUnsupportedError):
    pass

class SimCCallError(SimExpressionError):
    pass

class UnsupportedCCallError(SimCCallError, SimUnsupportedError):
    pass

class SimUninitializedAccessError(SimExpressionError):
    def __init__(self, expr_type, expr):
        SimExpressionError.__init__(self)
        self.expr_type = expr_type
        self.expr = expr

    def __repr__(self):
        return "SimUninitializedAccessError (expr %s is used as %s)" % (self.expr, self.expr_type)

    def __reduce__(self):
        return (SimUninitializedAccessError, (self.expr_type, self.expr))

#
# SimIRStmt errors
#

class SimStatementError(SimError):
    pass

class UnsupportedIRStmtError(SimStatementError, SimUnsupportedError):
    pass

class UnsupportedDirtyError(UnsupportedIRStmtError, SimUnsupportedError):
    pass

class SimMissingTempError(SimValueError, IndexError):
    pass

#
# Engine-related errors
#

class SimEngineError(SimError):
    pass

class SimIRSBError(SimEngineError):
    pass

class SimTranslationError(SimEngineError):
    pass

class SimProcedureError(SimEngineError):
    pass

class SimProcedureArgumentError(SimProcedureError):
    pass

class SimShadowStackError(SimProcedureError):
    pass

class SimFastPathError(SimEngineError):
    pass

class SimIRSBNoDecodeError(SimIRSBError):
    pass

class AngrUnsupportedSyscallError(AngrSyscallError, SimProcedureError, SimUnsupportedError):
    pass

UnsupportedSyscallError = AngrUnsupportedSyscallError

class SimReliftException(SimEngineError):
    def __init__(self, state):
        super(SimReliftException, self).__init__()
        self.state = state

#
# SimSlicer errors
#

class SimSlicerError(SimError):
    pass

#
# SimAction errors
#

class SimActionError(SimError):
    pass

#
# SimCC errors
#

class SimCCError(SimError):
    pass

#
# UCManager errors
#

class SimUCManagerError(SimError):
    pass

class SimUCManagerAllocationError(SimUCManagerError):
    pass

#
# SimUnicorn errors
#

class SimUnicornUnsupport(SimError):
    pass

class SimUnicornError(SimError):
    pass

class SimUnicornSymbolic(SimError):
    pass


#
# Call-stack Errors
#

class SimEmptyCallStackError(SimError):
    pass

#
# SimStateOptions Errors
#

class SimStateOptionsError(SimError):
    pass

#
# Errors that may be handled by exception handling
#

class SimException(SimError):
    pass

class SimSegfaultException(SimException, SimMemoryError):
    def __init__(self, addr, reason, original_addr=None):
        self.addr = addr
        self.reason = reason
        self.original_addr = original_addr
        super(SimSegfaultError, self).__init__('%#x (%s)' % (addr, reason))

    def __repr__(self):
        return 'SimSegfaultException(%#x (%s%s)' % (
            self.addr,
            self.reason,
            (', original %s' % self.original_addr.__repr__(max_depth=3)) if self.original_addr is not None else ''
        )

    def __reduce__(self):
        return (SimSegfaultException, (self.addr, self.reason, self.original_addr))

SimSegfaultError = SimSegfaultException

class SimZeroDivisionException(SimException, SimOperationError):
    pass


class AngrNoPluginError(AngrError):
    pass

#
# Concrete Targets Execution errors
#


class SimConcreteMemoryError(AngrError):
    pass


class SimConcreteRegisterError(AngrError):
    pass


class SimConcreteBreakpointError(AngrError):
    pass

#
# Decompiler errors
#

class UnsupportedNodeTypeError(AngrError, NotImplementedError):
    pass



from typing import Optional, List, Tuple, Any, Set, TYPE_CHECKING
import copy
import logging

import archinfo
from ailment import Stmt, Expr

from ...procedures.stubs.format_parser import FormatParser, FormatSpecifier
from ...errors import SimMemoryMissingError
from ...sim_type import SimTypeBottom, SimTypePointer, SimTypeChar, SimTypeInt
from ...calling_conventions import SimRegArg, SimStackArg, SimCC
from ...knowledge_plugins.key_definitions.constants import OP_BEFORE
from ...knowledge_plugins.key_definitions.definition import Definition
from .. import Analysis, register_analysis

if TYPE_CHECKING:
    from angr.calling_conventions import SimCC
    from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
    from angr.knowledge_plugins.key_definitions.live_definitions import LiveDefinitions


l = logging.getLogger(name=__name__)


class CallSiteMaker(Analysis):
    """
    Add calling convention, declaration, and args to a call site.
    """
    def __init__(self, block, reaching_definitions=None, stack_pointer_tracker=None, ail_manager=None):
        self.block = block

        self._reaching_definitions = reaching_definitions
        self._stack_pointer_tracker = stack_pointer_tracker
        self._ail_manager = ail_manager

        self.result_block = None
        self.stack_arg_offsets: Optional[Set[Tuple[int,int]]] = None  # ins_addr, stack_offset

        self._analyze()

    def _analyze(self):

        if not self.block.statements:
            return

        last_stmt = self.block.statements[-1]

        if not type(last_stmt) is Stmt.Call:
            self.result_block = self.block
            return

        cc = None
        prototype = None
        args = None
        stack_arg_locs: List[SimStackArg] = [ ]
        stackarg_sp_diff = 0

        target = self._get_call_target(last_stmt)
        if target is not None and target in self.kb.functions:
            # function-specific logic when the calling target is known
            func = self.kb.functions[target]
            if func.prototype is None:
                func.find_declaration()
            cc = func.calling_convention
            prototype = func.prototype

            args = [ ]
            arg_locs = None
            if cc is None:
                l.warning('%s has an unknown calling convention.', repr(func))
            else:
                stackarg_sp_diff = func.calling_convention.STACKARG_SP_DIFF
                if prototype is not None:
                    # Make arguments
                    arg_locs = cc.arg_locs(prototype)
                    if prototype.variadic:
                        # determine the number of variadic arguments
                        variadic_args = self._determine_variadic_arguments(func, cc, last_stmt)
                        if variadic_args:
                            callsite_ty = copy.copy(prototype)
                            callsite_ty.args = list(callsite_ty.args)
                            for i in range(variadic_args):
                                callsite_ty.args.append(SimTypeInt().with_arch(self.project.arch))
                            arg_locs = cc.arg_locs(callsite_ty)

            if arg_locs is not None:
                for arg_loc in arg_locs:
                    if type(arg_loc) is SimRegArg:
                        size = arg_loc.size
                        offset = arg_loc.check_offset(cc.arch)
                        args.append(Expr.Register(self._atom_idx(), None, offset, size * 8,
                                                  reg_name=arg_loc.reg_name))
                    elif type(arg_loc) is SimStackArg:

                        stack_arg_locs.append(arg_loc)
                        _, the_arg = self._resolve_stack_argument(last_stmt, arg_loc)

                        if the_arg is not None:
                            args.append(the_arg)
                        else:
                            args.append(None)

                    else:
                        raise NotImplementedError('Not implemented yet.')

        # Remove the old call statement
        new_stmts = self.block.statements[:-1]

        # remove the statement that stores the return address
        if self.project.arch.call_pushes_ret:
            # check if the last statement is storing the return address onto the top of the stack
            if len(new_stmts) >= 1:
                the_stmt = new_stmts[-1]
                if isinstance(the_stmt, Stmt.Store) and isinstance(the_stmt.data, Expr.Const):
                    if isinstance(the_stmt.addr, Expr.StackBaseOffset) and \
                            the_stmt.data.value == self.block.addr + self.block.original_size:
                        # yes it is!
                        new_stmts = new_stmts[:-1]
        else:
            # if there is an lr register...
            lr_offset = None
            if archinfo.arch_arm.is_arm_arch(self.project.arch) or self.project.arch.name in {'PPC32', 'PPC64'}:
                lr_offset = self.project.arch.registers['lr'][0]
            elif self.project.arch.name in {'MIPS32', 'MIPS64'}:
                lr_offset = self.project.arch.registers['ra'][0]
            if lr_offset is not None:
                # remove the assignment to the lr register
                if len(new_stmts) >= 1:
                    the_stmt = new_stmts[-1]
                    if (isinstance(the_stmt, Stmt.Assignment)
                            and isinstance(the_stmt.dst, Expr.Register)
                            and the_stmt.dst.reg_offset == lr_offset
                    ):
                        # found it
                        new_stmts = new_stmts[:-1]

        # calculate stack offsets for arguments that are put on the stack. these offsets will be consumed by
        # simplification steps in the future, which may decide to remove statements that store arguments on the stack.
        if stack_arg_locs:
            sp_offset = self._stack_pointer_tracker.offset_before(last_stmt.ins_addr, self.project.arch.sp_offset)
            if sp_offset is None:
                l.warning("Failed to calculate the stack pointer offset at pc %#x. You may find redundant Store "
                          "statements.", last_stmt.ins_addr)
                self.stack_arg_offsets = None
            else:
                self.stack_arg_offsets = set(
                    (last_stmt.ins_addr, sp_offset + arg.stack_offset - stackarg_sp_diff) for arg in stack_arg_locs
                )

        ret_expr = last_stmt.ret_expr
        # if ret_expr is None, it means in previous steps (such as during AIL simplification) we have deemed the return
        # value of this call statement as useless and is removed.

        if ret_expr is not None \
                and prototype is not None \
                and prototype.returnty is not None \
                and not isinstance(prototype.returnty, SimTypeBottom):
            # try to narrow the return expression if needed
            ret_type_bits = prototype.returnty.with_arch(self.project.arch).size
            if ret_expr.bits > ret_type_bits:
                ret_expr = ret_expr.copy()
                ret_expr.bits = ret_type_bits

        new_stmts.append(Stmt.Call(last_stmt.idx, last_stmt.target,
                                   calling_convention=cc,
                                   prototype=prototype,
                                   args=args,
                                   ret_expr=ret_expr,
                                   **last_stmt.tags,
                                   ))

        new_block = self.block.copy()
        new_block.statements = new_stmts

        self.result_block = new_block

    def _find_variable_from_definition(self, def_):
        """

        :param Definition def_: The reaching definition of a variable.
        :return:                The variable that is defined.
        """

        if def_.codeloc.block_addr != self.block.addr:
            l.warning("TODO: The definition comes from a different block %#x.", def_.codeloc.block_addr)
            return None

        stmt = self.block.statements[def_.codeloc.stmt_idx]
        if type(stmt) is Stmt.Assignment:
            return stmt.dst
        elif type(stmt) is Stmt.Store:
            return stmt.addr
        else:
            l.warning("TODO: Unsupported statement type %s for definitions.", type(stmt))
            return None

    def _resolve_register_argument(self, call_stmt, arg_loc) -> Tuple:

        size = arg_loc.size
        offset = arg_loc.check_offset(self.project.arch)

        if self._reaching_definitions is not None:
            # Find its definition
            ins_addr = call_stmt.tags['ins_addr']
            try:
                rd: 'LiveDefinitions' = self._reaching_definitions.get_reaching_definitions_by_insn(ins_addr, OP_BEFORE)
            except KeyError:
                return None, None

            try:
                vs: 'MultiValues' = rd.register_definitions.load(offset, size=size)
            except SimMemoryMissingError:
                return None, None
            values_and_defs_ = set()
            for values in vs.values():
                for value in values:
                    if value.concrete:
                        concrete_value = value._model_concrete.value
                    else:
                        concrete_value = None
                    for def_ in rd.extract_defs(value):
                        values_and_defs_.add((concrete_value, def_))

            if not values_and_defs_:
                l.warning("Did not find any reaching definition for register %s at instruction %x.",
                          arg_loc, ins_addr)
            elif len(values_and_defs_) > 1:
                l.warning("TODO: More than one reaching definition are found at instruction %x.",
                          ins_addr)
            else:
                # Find the definition
                # FIXME: Multiple definitions - we need phi nodes
                value, def_ = next(iter(values_and_defs_))  # type:Definition
                variable = self._find_variable_from_definition(def_)
                return value, variable

        return None, None

    def _resolve_stack_argument(self, call_stmt, arg_loc) -> Tuple[Any,Any]:  # pylint:disable=unused-argument

        size = arg_loc.size
        offset = arg_loc.stack_offset
        if self.project.arch.call_pushes_ret:
            # adjust the offset
            offset -= self.project.arch.bytes

        # TODO: Support extracting values

        return None, Expr.Load(
            self._atom_idx(),
            Expr.Register(self._atom_idx(), None, self.project.arch.sp_offset, self.project.arch.bits) +
            Expr.Const(self._atom_idx(), None, offset, self.project.arch.bits),
            size,
            self.project.arch.memory_endness,
            func_arg=True,
        )

    @staticmethod
    def _get_call_target(stmt):
        """

        :param Stmt.Call stmt:
        :return:
        """

        if type(stmt.target) is Expr.Const:
            return stmt.target.value

        return None

    def _load_string(self, addr: int) -> bytes:

        s = b""
        while True:
            try:
                chunk = self.project.loader.memory.load(addr, 8)
                addr += 8
            except KeyError:
                return s

            if b"\x00" in chunk:
                # found a null byte
                s += chunk[:chunk.index(b"\x00")]
                return s
            s += chunk
            if len(s) > 2048:
                break

        return s

    def _determine_variadic_arguments(self, func, cc: 'SimCC', call_stmt) -> Optional[int]:
        if "printf" in func.name or "scanf" in func.name:
            return self._determine_variadic_arguments_for_format_strings(func, cc, call_stmt)
        return None

    def _determine_variadic_arguments_for_format_strings(self, func, cc: 'SimCC', call_stmt) -> Optional[int]:
        proto = func.prototype
        if proto is None:
            # TODO: Support cases where prototypes are not available
            return None

        #
        # get the format string
        #

        potential_fmt_args: List[int] = [ ]
        for idx, arg in enumerate(proto.args):
            if isinstance(arg, SimTypePointer) and isinstance(arg.pts_to, SimTypeChar):
                # find a char*
                # we assume this is the format string
                potential_fmt_args.append(idx)

        if not potential_fmt_args:
            return None

        fmt_str = None
        min_arg_count = (max(potential_fmt_args) + 1)
        arg_locs = cc.arg_locs(SimCC.guess_prototype([0]*min_arg_count, proto))

        for fmt_arg_idx in potential_fmt_args:
            arg_loc = arg_locs[fmt_arg_idx]

            if isinstance(arg_loc, SimRegArg):
                value, _ = self._resolve_register_argument(call_stmt, arg_loc)
            elif isinstance(arg_loc, SimStackArg):
                value, _ = self._resolve_stack_argument(call_stmt, arg_loc)
            else:
                # Unexpected type of argument
                l.warning("Unexpected type of argument type %s.", arg_loc.__class__)
                return None

            if isinstance(value, int):
                fmt_str = self._load_string(value)
                if fmt_str:
                    break

        if not fmt_str:
            return None

        #
        # parse the format string
        #

        parser = FormatParser(project=self.project)
        fmt_str_list = [ bytes([b]) for b in fmt_str ]
        components = parser.extract_components(fmt_str_list)

        specifiers = [component for component in components if isinstance(component, FormatSpecifier)]
        if not specifiers:
            return None
        return len(specifiers)

    def _atom_idx(self) -> Optional[int]:
        return self._ail_manager.next_atom() if self._ail_manager is not None else None


register_analysis(CallSiteMaker, 'AILCallSiteMaker')



import angr

######################################
# read
######################################

class read(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, fd, dst, length):
        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1

        return simfd.read(dst, length)



from typing import Dict, Tuple, Set, Optional, TYPE_CHECKING

from .uses import Uses
from .live_definitions import LiveDefinitions

if TYPE_CHECKING:
    from angr.knowledge_plugins.key_definitions.definition import Definition


# TODO: Make ReachingDefinitionsModel serializable
class ReachingDefinitionsModel:
    """
    Models the definitions, uses, and memory of a ReachingDefinitionState object
    """
    def __init__(self, func_addr: Optional[int]=None):
        self.func_addr = func_addr  # do not use. only for pretty-printing
        self.observed_results: Dict[Tuple[str, int, int], LiveDefinitions] = {}
        self.all_definitions: Set['Definition'] = set()
        self.all_uses = Uses()

    def __repr__(self):
        return "<RDModel{} with {} observations>".format(
            "[func %#x]" if self.func_addr is not None else "",
            len(self.observed_results),
        )

    def copy(self) -> "ReachingDefinitionsModel":
        new = ReachingDefinitionsModel(self.func_addr)
        new.observed_results = self.observed_results.copy()
        new.all_definitions = self.all_definitions.copy()
        new.all_uses = self.all_uses.copy()
        return new

    def merge(self, model: 'ReachingDefinitionsModel'):
        for k, v in model.observed_results.items():
            if k not in self.observed_results:
                self.observed_results[k] = v
            else:
                merged, merge_occured = self.observed_results[k].merge(v)
                if merge_occured:
                    self.observed_results[k] = merged
        self.all_definitions.union(model.all_definitions)
        self.all_uses.merge(model.all_uses)



from typing import Tuple, Optional

from ailment.expression import BinaryOp, Const, Expression, Convert
from ailment.statement import Call

from .base import PeepholeOptimizationExprBase


class Bswap(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Simplifying bswap_16()"
    expr_classes = (BinaryOp, )  # all expressions are allowed

    def optimize(self, expr: BinaryOp):

        # bswap_16
        #   And(
        #     (
        #       ((Conv(16->32, A) << 0x8<8>) & 0xff00ff00<32>) |
        #       ((Conv(16->32, A) >> 0x8<8>) & 0xff00ff<32>)
        #     ),
        #     0xffff<32>
        #   )
        if expr.op == "And" and len(expr.operands) == 2 and isinstance(expr.operands[1], Const) \
                and expr.operands[1].value == 0xffff:
            inner = expr.operands[0]
            if isinstance(inner, BinaryOp) and inner.op == "Or" and len(inner.operands) == 2:
                or_first, or_second = inner.operands[0], inner.operands[1]
                if isinstance(or_first, BinaryOp) and or_first.op == "And" and len(or_first.operands) == 2 and \
                        isinstance(or_second, BinaryOp) and or_second.op == "And" and len(or_second.operands) == 2:
                    r, the_expr = self._match_inner(or_first, or_second)
                    if r:
                        return Call(expr.idx, "__builtin_bswap16", args=[the_expr], bits=expr.bits, **expr.tags)

                    r, the_expr = self._match_inner(or_second, or_first)
                    if r:
                        return Call(expr.idx, "__builtin_bswap16", args=[the_expr], bits=expr.bits, **expr.tags)

                    return None

        return None

    def _match_inner(self, or_first: BinaryOp, or_second: BinaryOp) -> Tuple[bool,Optional[Expression]]:
        if isinstance(or_first.operands[1], Const) and or_first.operands[1].value == 0xff00ff00:
            if isinstance(or_second.operands[1], Const) and or_second.operands[1].value == 0x00ff00ff:
                inner_first = or_first.operands[0]
                inner_second = or_second.operands[0]
                if isinstance(inner_first, BinaryOp) and inner_first.op == "Shl" and \
                        isinstance(inner_first.operands[1], Const) and inner_first.operands[1].value == 8:
                    if isinstance(inner_second, BinaryOp) and inner_second.op == "Shr" and \
                            isinstance(inner_second.operands[1], Const) and inner_second.operands[1].value == 8:

                        if isinstance(inner_first.operands[0], Convert):
                            conv: Convert = inner_first.operands[0]
                            if conv.from_bits == 16 and conv.to_bits == 32:
                                the_expr_1 = conv.operand
                                if isinstance(inner_second.operands[0], Convert) and \
                                        inner_second.operands[0].from_bits == 16 and \
                                        inner_second.operands[0].to_bits == 32:
                                    the_expr_2 = inner_second.operands[0].operand
                                    if the_expr_1.likes(the_expr_2):
                                        return True, the_expr_1
        return False, None



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("wmvcore.dll")
prototypes = \
    {
        # 
        'WMIsContentProtected': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszFileName", "pfIsProtected"]),
        # 
        'WMCreateWriter': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypePointer(SimTypeBottom(label="IWMWriter"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pUnkCert", "ppWriter"]),
        # 
        'WMCreateReader': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWMReader"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pUnkCert", "dwRights", "ppReader"]),
        # 
        'WMCreateSyncReader': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWMSyncReader"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pUnkCert", "dwRights", "ppSyncReader"]),
        # 
        'WMCreateEditor': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWMMetadataEditor"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppEditor"]),
        # 
        'WMCreateIndexer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWMIndexer"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppIndexer"]),
        # 
        'WMCreateBackupRestorer': SimTypeFunction([SimTypeBottom(label="IUnknown"), SimTypePointer(SimTypeBottom(label="IWMLicenseBackup"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pCallback", "ppBackup"]),
        # 
        'WMCreateProfileManager': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWMProfileManager"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppProfileManager"]),
        # 
        'WMCreateWriterFileSink': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWMWriterFileSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppSink"]),
        # 
        'WMCreateWriterNetworkSink': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWMWriterNetworkSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppSink"]),
        # 
        'WMCreateWriterPushSink': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWMWriterPushSink"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppSink"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=abstract-method,arguments-differ
import logging
from typing import Optional, List, Set, Tuple, Union, Callable

from angr.utils.dynamic_dictlist import DynamicDictList
from .....storage.memory_object import SimMemoryObject, SimLabeledMemoryObject
from . import PageBase
from .cooperation import MemoryObjectSetMixin


l = logging.getLogger(name=__name__)

_MOTYPE = Union[SimMemoryObject, SimLabeledMemoryObject]


class MVListPage(
    MemoryObjectSetMixin,
    PageBase,
):
    """
    MVListPage allows storing multiple values at the same location, thus allowing weak updates.

    Each store() may take a value or multiple values, and a "weak" parameter to specify if this store is a weak update
    or not.
    Each load() returns an iterator of all values stored at that location.
    """
    def __init__(self, memory=None, content=None, sinkhole=None, mo_cmp=None, **kwargs):
        super().__init__(**kwargs)

        self.content: DynamicDictList[Optional[Union[_MOTYPE,Set[_MOTYPE]]]] = \
            DynamicDictList(max_size=memory.page_size, content=content) if content is not None else None
        self.stored_offset = set()
        self._mo_cmp: Optional[Callable] = mo_cmp

        if self.content is None:
            if memory is not None:
                self.content: DynamicDictList[Optional[Union[_MOTYPE,Set[_MOTYPE]]]] = \
                    DynamicDictList(max_size=memory.page_size)

        self.sinkhole: Optional[_MOTYPE] = sinkhole

    def copy(self, memo) -> 'MVListPage':
        o = super().copy(memo)
        o.content = DynamicDictList(max_size=self.content.max_size, content=self.content)
        o.sinkhole = self.sinkhole
        o.stored_offset = self.stored_offset.copy()
        o._mo_cmp = self._mo_cmp
        return o

    def load(self, addr, size=None, endness=None, page_addr=None, memory=None, cooperate=False,
             **kwargs) -> List[Tuple[int,_MOTYPE]]:
        result = [ ]
        last_seen = ...  # ;)

        # loop over the loading range. accumulate a result for each byte, but collapse results from adjacent bytes
        # using the same memory object
        for subaddr in range(addr, addr + size):
            items = self.content[subaddr]
            if items is None:
                items = { self.sinkhole } if self.sinkhole is not None else None
            if items != last_seen:
                if last_seen is None:
                    self._fill(result, subaddr, page_addr, endness, memory, **kwargs)
                result.append((subaddr + page_addr, items))
                last_seen = items

        if last_seen is None:
            self._fill(result, addr + size, page_addr, endness, memory, **kwargs)

        if not cooperate:
            result = self._force_load_cooperation(result, size, endness, memory=memory, **kwargs)
        return result

    def _fill(self, result, addr, page_addr, endness, memory, **kwargs):
        """
        Small utility function for behavior which is duplicated in load

        mutates result to generate a new memory object and replace the last entry in it, which is None. Then, it will
        insert the new memory object into self.content.
        """
        global_end_addr = addr + page_addr
        global_start_addr = result[-1][0]
        size = global_end_addr - global_start_addr
        new_ast = self._default_value(global_start_addr,
                                      size,
                                      name='%s_%x' % (memory.id, global_start_addr),
                                      key=(self.category, global_start_addr),
                                      memory=memory,
                                      **kwargs)
        new_item = SimMemoryObject(new_ast, global_start_addr, endness=endness,
                                   byte_width=memory.state.arch.byte_width if memory is not None else 8)
        subaddr_start = global_start_addr - page_addr
        for subaddr in range(subaddr_start, addr):
            self.content[subaddr] = { new_item }
            self.stored_offset.add(subaddr)
        result[-1] = (global_start_addr, new_item)

    def store(self, addr, data, size=None, endness=None, memory=None, cooperate=False, weak=False, **kwargs):
        super().store(addr, data, size=size, endness=endness, memory=memory, cooperate=cooperate, weak=weak, **kwargs)

        if not cooperate:
            data = self._force_store_cooperation(addr, data, size, endness, memory=memory, **kwargs)

        data: Set[_MOTYPE]

        if size == len(self.content) and addr == 0 and len(data) == 1:
            self.sinkhole = next(iter(data))
            self.content = DynamicDictList(max_size=len(self.content))
            self.stored_offset = set()
        else:
            if not weak:
                if len(data) == 1:
                    # unpack
                    data: _MOTYPE = next(iter(data))
                for subaddr in range(addr, addr + size):
                    self.content[subaddr] = data
                    self.stored_offset.add(subaddr)
            else:
                for subaddr in range(addr, addr + size):
                    if self.content[subaddr] is None:
                        self.content[subaddr] = data
                    elif type(self.content[subaddr]) is set:
                        self.content[subaddr] |= data
                    else:
                        self.content[subaddr] = {self.content[subaddr]} | data
                    self.stored_offset.add(subaddr)

    def erase(self, addr, size=None, **kwargs) -> None:
        for off in range(size):
            self.content[addr + off] = None

    def merge(self, others: List['MVListPage'], merge_conditions, common_ancestor=None, page_addr: int = None,
              memory=None, changed_offsets: Optional[Set[int]]=None):

        if changed_offsets is None:
            changed_offsets = set()
            for other in others:
                changed_offsets |= self.changed_bytes(other, page_addr)

        all_pages: List['MVListPage'] = [self] + others
        if merge_conditions is None:
            merge_conditions = [None] * len(all_pages)

        merged_to = None
        merged_objects = set()
        merged_offsets = set()
        for b in sorted(changed_offsets):
            if merged_to is not None and not b >= merged_to:
                l.info("merged_to = %d ... already merged byte 0x%x", merged_to, b)
                continue
            l.debug("... on byte 0x%x", b)

            memory_objects = []
            unconstrained_in = []

            # first get a list of all memory objects at that location, and
            # all memories that don't have those bytes
            for sm, fv in zip(all_pages, merge_conditions):
                if sm._contains(b, page_addr):
                    l.info("... present in %s", fv)
                    for mo in sm.content_gen(b):
                        if mo.includes(page_addr + b):
                            memory_objects.append((mo, fv))
                else:
                    l.info("... not present in %s", fv)
                    unconstrained_in.append((sm, fv))

            if not memory_objects:
                continue

            mos = set(mo for mo, _ in memory_objects)
            mo_bases = set(mo.base for mo, _ in memory_objects)
            mo_lengths = set(mo.length for mo, _ in memory_objects)
            endnesses = set(mo.endness for mo in mos)

            if not unconstrained_in and not (mos - merged_objects):  # pylint:disable=superfluous-parens
                continue

            # first, optimize the case where we are dealing with the same-sized memory objects
            if len(mo_bases) == 1 and len(mo_lengths) == 1 and not unconstrained_in and len(endnesses) == 1:
                the_endness = next(iter(endnesses))
                to_merge = [(mo.object, fv) for mo, fv in memory_objects]

                # Update `merged_to`
                mo_base = list(mo_bases)[0]
                mo_length = memory_objects[0][0].length
                size = min(mo_length - (page_addr + b - mo_base), len(self.content) - b)
                merged_to = b + size

                merged_val = self._merge_values(to_merge, mo_length, memory=memory)
                if merged_val is None:
                    # merge_values() determines that we should not attempt to merge this value
                    continue

                # do the replacement
                # TODO: Implement in-place replacement instead of calling store()
                # new_object = self._replace_memory_object(our_mo, merged_val, page_addr, memory.page_size)

                first_value = True
                for v in merged_val:
                    self.store(b,
                               { SimMemoryObject(v, mo_base, endness=the_endness) },
                               size=size,
                               cooperate=True,
                               weak=not first_value,
                               )
                    first_value = False

                merged_offsets.add(b)

            else:
                # get the size that we can merge easily. This is the minimum of
                # the size of all memory objects and unallocated spaces.
                min_size = min([mo.length - (b + page_addr - mo.base)
                                for mo, _ in memory_objects] + [len(self.content) - b])
                for um, _ in unconstrained_in:
                    for i in range(0, min_size):
                        if um._contains(b + i, page_addr):
                            min_size = i
                            break
                merged_to = b + min_size
                l.info("... determined minimum size of %d", min_size)

                # Now, we have the minimum size. We'll extract/create expressions of that
                # size and merge them
                extracted = [(mo.bytes_at(page_addr + b, min_size), fv) for mo, fv in
                             memory_objects] if min_size != 0 else []
                if not memory.skip_missing_values_during_merging:
                    created = [
                        (self._default_value(None, min_size, name="merge_uc_%s_%x" % (uc.id, b), memory=memory),
                         fv) for
                        uc, fv in unconstrained_in
                    ]
                    to_merge = extracted + created
                else:
                    to_merge = extracted

                merged_val = self._merge_values(to_merge, min_size, memory=memory)
                if merged_val is None:
                    continue

                first_value = True
                for v in merged_val:
                    self.store(b,
                               { SimMemoryObject(v, page_addr + b, endness='Iend_BE') },
                               size=min_size,
                               endness='Iend_BE',
                               cooperate=True,
                               weak=not first_value,
                               )  # do not convert endianness again
                    first_value = False
                merged_offsets.add(b)

        self.stored_offset |= merged_offsets
        return merged_offsets

    def changed_bytes(self, other: 'MVListPage', page_addr: int = None):
        candidates: Set[int] = super().changed_bytes(other)
        if candidates is None:
            candidates: Set[int] = set()
            # resort to the slower solution
            if self.sinkhole is None:
                candidates |= self.stored_offset
            else:
                for i in range(len(self.content)):
                    if self._contains(i, page_addr):
                        candidates.add(i)

            if other.sinkhole is None:
                candidates |= other.stored_offset
            else:
                for i in range(len(other.content)):
                    if other._contains(i, page_addr):
                        candidates.add(i)

        byte_width = 8  # TODO: Introduce self.state if we want to use self.state.arch.byte_width
        differences: Set[int] = set()
        for c in candidates:
            s_contains = self._contains(c, page_addr)
            o_contains = other._contains(c, page_addr)
            if not s_contains and o_contains:
                differences.add(c)
            elif s_contains and not o_contains:
                differences.add(c)
            else:
                if self.content[c] is None:
                    if self.sinkhole is not None:
                        self.content[c] = SimMemoryObject(self.sinkhole.bytes_at(page_addr + c, 1), page_addr + c,
                                                          byte_width=byte_width, endness='Iend_BE')
                if other.content[c] is None:
                    if other.sinkhole is not None:
                        other.content[c] = SimMemoryObject(other.sinkhole.bytes_at(page_addr + c, 1), page_addr + c,
                                                           byte_width=byte_width, endness='Iend_BE')
                if s_contains and self.content[c] != other.content[c]:
                    same = None
                    if self._mo_cmp is not None:
                        same = self._mo_cmp(self.content[c], other.content[c], page_addr + c, 1)
                    if same is None:
                        # Try to see if the bytes are equal
                        self_bytes = { mo.bytes_at(page_addr + c, 1) for mo in self.content_gen(c) }
                        other_bytes = { mo.bytes_at(page_addr + c, 1) for mo in other.content_gen(c) }
                        same = self_bytes == other_bytes

                    if same is False:
                        differences.add(c)
                else:
                    # this means the byte is in neither memory
                    pass

        return differences

    def content_gen(self, index):
        if self.content[index] is None:
            return
        elif type(self.content[index]) is set:
            yield from self.content[index]
        else:
            yield self.content[index]

    def _contains(self, off: int, page_addr: int):
        if off >= len(self.content):
            return False
        if self.content[off] is not None:
            return True
        if self.sinkhole is None:
            return False
        return self.sinkhole.includes(page_addr + off)

    def _replace_mo(self, old_mo: SimMemoryObject, new_mo: SimMemoryObject, page_addr: int,
                    page_size: int) -> SimMemoryObject:
        if self.sinkhole is old_mo:
            self.sinkhole = new_mo
        else:
            start, end = self._resolve_range(old_mo, page_addr, page_size)
            for i in range(start, end):
                s = { new_mo }
                if self.content[i - page_addr] is old_mo:
                    self.content[i - page_addr] = s
        return new_mo

    @staticmethod
    def _resolve_range(mo: SimMemoryObject, page_addr: int, page_size) -> Tuple[int, int]:
        start = max(mo.base, page_addr)
        end = min(mo.last_addr + 1, page_addr + page_size)
        if end <= start:
            l.warning("Nothing left of the memory object to store in SimPage.")
        return start, end

    def _get_objects(self, start: int, page_addr: int) -> Optional[List[SimMemoryObject]]:
        mos = self.content[start]
        if mos is None:
            return None
        lst = [ ]
        if type(mos) is set:
            for mo in mos:
                if mo.includes(start + page_addr):
                    lst.append(mo)
        else:
            lst.append(mos)
        if lst:
            return lst
        return None



from typing import Optional

from ailment.expression import BinaryOp, UnaryOp
from angr.project import Project
from angr.knowledge_base import KnowledgeBase


class PeepholeOptimizationStmtBase:
    """
    The base class for all peephole optimizations that are applied on AIL statements.
    """

    __slots__ = ('project', 'kb', 'func_addr', )
    project: Project
    kb: KnowledgeBase
    func_addr: Optional[int]

    NAME = "Peephole Optimization - Statement"
    DESCRIPTION = "Peephole Optimization - Statement"
    stmt_classes = None

    def __init__(self, project: Project, kb: KnowledgeBase, func_addr: Optional[int] = None):
        self.project = project
        self.kb = kb
        self.func_addr = func_addr

    def optimize(self, stmt):
        raise NotImplementedError("_optimize() is not implemented.")


class PeepholeOptimizationExprBase:
    """
    The base class for all peephole optimizations that are applied on AIL expressions.
    """

    __slots__ = ('project', 'kb', 'func_addr', )
    project: Project
    kb: KnowledgeBase
    func_addr: Optional[int]

    NAME = "Peephole Optimization - Expression"
    DESCRIPTION = "Peephole Optimization - Expression"
    expr_classes = None

    def __init__(self, project: Project, kb: KnowledgeBase, func_addr: Optional[int] = None):
        self.project = project
        self.kb = kb
        self.func_addr = func_addr

    def optimize(self, expr):
        raise NotImplementedError("_optimize() is not implemented.")

    #
    # Util methods
    #

    @staticmethod
    def is_bool_expr(ail_expr):

        if isinstance(ail_expr, BinaryOp):
            if ail_expr.op in {'CmpEQ', 'CmpNE', 'CmpLT', 'CmpLE', 'CmpGT', 'CmpGE', 'CmpLTs', 'CmpLEs', 'CmpGTs',
                               'CmpGEs'}:
                return True
        if isinstance(ail_expr, UnaryOp) and ail_expr.op == 'Not':
            return True
        return False



import angr

######################################
# fgetc
######################################


class fgetc(angr.SimProcedure):
    # pylint:disable=arguments-differ
    def run(self, stream, simfd=None):
        if simfd is None:
            fileno = angr.SIM_PROCEDURES['posix']['fileno']
            fd = self.inline_call(fileno, stream).ret_expr
            simfd = self.state.posix.get_fd(fd)

        if simfd is None:
            return -1

        data, real_length, = simfd.read_data(1)
        return self.state.solver.If(real_length == 0, -1, data.zero_extend(self.arch.sizeof['int'] - 8))

getc = fgetc
fgetc_unlocked = fgetc
getc_unlocked = fgetc



import math

from ailment.expression import Convert, BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ConvConstMullAShift(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Conv(64->32, (N * a) >> M) => a / N1"
    expr_classes = (Convert, )

    def optimize(self, expr: Convert):

        if expr.from_bits == 64 and expr.to_bits == 32:
            if isinstance(expr.operand, BinaryOp) and expr.operand.op == "Shr" \
                    and isinstance(expr.operand.operands[1], Const):
                # (N * a) >> M  ==>  a / N1
                inner = expr.operand.operands[0]
                if isinstance(inner, BinaryOp) and inner.op == "Mull" and isinstance(inner.operands[0], Const):
                    C = inner.operands[0].value
                    X = inner.operands[1]
                    V = expr.operand.operands[1].value
                    ndigits = 5 if V == 32 else 6
                    divisor = self._check_divisor(pow(2, V), C, ndigits)
                    if divisor is not None:
                        new_const = Const(None, None, divisor, V)
                        new_div = BinaryOp(inner.idx, 'Div', [X, new_const], inner.signed, **inner.tags)
                        return new_div

            elif isinstance(expr.operand, BinaryOp) and expr.operand.op in {"Add", "Sub"}:
                expr0, expr1 = expr.operand.operands
                if isinstance(expr0, BinaryOp) \
                        and expr0.op in {"Shr", "Sar"} \
                        and isinstance(expr0.operands[1], Const) \
                        and isinstance(expr1, BinaryOp) \
                        and expr1.op in {"Shr", "Sar"} \
                        and isinstance(expr1.operands[1], Const):
                    if isinstance(expr0.operands[0], BinaryOp) \
                            and expr0.operands[0].op in {"Mull", "Mul"} \
                            and isinstance(expr0.operands[0].operands[1], Const):
                        a0 = expr0.operands[0].operands[0]
                        a1 = expr1.operands[0]
                        if a0 == a1:
                            # (a * x >> M1) +/- (a >> M2)  ==>  a / N
                            C = expr0.operands[0].operands[1].value
                            X = a0
                            V = expr0.operands[1].value
                            ndigits = 5 if V == 32 else 6
                            divisor = self._check_divisor(pow(2, V), C, ndigits)
                            if divisor is not None:
                                new_const = Const(None, None, divisor, V)
                                new_div = BinaryOp(expr0.operands[0].idx, 'Div', [X, new_const],
                                                   expr0.operands[0].signed, **expr0.operands[0].tags)
                                # we cannot drop the convert
                                new_div = Convert(expr.idx, expr.from_bits, expr.to_bits, expr.is_signed,
                                                  new_div, **expr.tags)
                                return new_div

    @staticmethod
    def _check_divisor(a, b, ndigits=6):
        divisor_1 = 1 + (a//b)
        divisor_2 = int(round(a/float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None



from typing import Optional, Any, Callable, TYPE_CHECKING

from ailment import Block, Stmt, Expr

from ...code_location import CodeLocation
from ..decompiler.ailblock_walker import AILBlockWalker

if TYPE_CHECKING:
    from .propagator import PropagatorAILState


class OutdatedDefinitionWalker(AILBlockWalker):
    """
    Walks an AIL expression to find outdated definitions.
    """
    def __init__(self, expr, expr_defat: CodeLocation, state: 'PropagatorAILState',
                 avoid: Optional[Expr.Expression]=None, extract_offset_to_sp: Callable=None):
        super().__init__()
        self.expr = expr
        self.expr_defat = expr_defat
        self.state = state
        self.avoid = avoid
        self.extract_offset_to_sp = extract_offset_to_sp
        self.expr_handlers[Expr.Register] = self._handle_Register
        self.expr_handlers[Expr.Load] = self._handle_Load
        self.expr_handlers[Expr.Tmp] = self._handle_Tmp
        self.expr_handlers[Expr.VEXCCallExpression] = self._handle_VEXCCallExpression
        self.out_dated = False

    # pylint:disable=unused-argument
    def _handle_Tmp(self, expr_idx: int, tmp_expr: Expr.Tmp, stmt_idx: int, stmt: Stmt.Assignment,
                    block: Optional[Block]):
        if self.avoid is not None and tmp_expr.likes(self.avoid):
            self.out_dated = True

    # pylint:disable=unused-argument
    def _handle_Register(self, expr_idx: int, reg_expr: Expr.Register, stmt_idx: int, stmt: Stmt.Assignment,
                         block: Optional[Block]):
        if self.avoid is not None and isinstance(self.avoid, Expr.Register) and (
                reg_expr.likes(self.avoid) or self._reg_overlap(reg_expr, self.avoid)):
            self.out_dated = True
        else:
            v = self.state.load_register(reg_expr)
            if v is not None:
                for detail in v.offset_and_details.values():
                    if detail.expr is None:
                        self.out_dated = True
                        break
                    if isinstance(detail.expr, Expr.Expression) and detail.expr.has_atom(reg_expr, identity=False):
                        self.out_dated = True
                        break
                    if isinstance(detail.expr, Expr.TaggedObject):
                        if not (detail.def_at == self.expr_defat or
                                self._check_store_precedes_load(detail.def_at, self.expr_defat)):
                            self.out_dated = True
                            break

    @staticmethod
    def _check_store_precedes_load(store_defat: Optional[CodeLocation], load_defat: Optional[CodeLocation]) -> bool:
        """
        Check if store precedes load based on their AIL statement IDs.
        """
        if store_defat is None or load_defat is None:
            return True
        return store_defat.block_addr == load_defat.block_addr and store_defat.stmt_idx <= load_defat.stmt_idx

    @staticmethod
    def _check_global_store_conflicts_load(store_block_addr: int, store_stmt_idx: int, addr: Any, store: Stmt.Store,
                                           load_block_addr: int, load_stmt_idx: int, load: Expr.Load) -> bool:
        """
        Check if the load may conflict with any existing stores that happened in the past.

        :param addr:
        :param size:
        :param store:
        :param load:
        :return:
        """
        if store_block_addr == load_block_addr and store_stmt_idx >= load_stmt_idx:
            written = set()
            if isinstance(addr, Expr.Const) and isinstance(store.size, int):
                written |= { addr.value + i for i in range(store.size) }
            else:
                return False  # FIXME: This is unsafe
            if isinstance(load.addr, Expr.Const) and isinstance(load.size, int):
                read = { load.addr.value + i for i in range(load.size) }
                return bool(written.intersection(read))
            return True
        return False

    def _handle_Load(self, expr_idx: int, expr: Expr.Load, stmt_idx: int, stmt: Stmt.Statement, block: Optional[Block]):
        if self.avoid is not None and (expr == self.avoid or expr.addr == self.avoid):  # pylint:disable=consider-using-in
            self.out_dated = True
        elif isinstance(expr.addr, Expr.StackBaseOffset):
            sp_offset = self.extract_offset_to_sp(expr.addr)
            if sp_offset is not None:
                # if expr.addr can be resolved to a concrete stack offset, perform the read and detect if the variables
                # on the stack are updated after their values are loaded into self.expr
                curr_stackvar = self.state.load_stack_variable(sp_offset, expr.size, endness=expr.endness)
                if curr_stackvar is None:
                    # the variable does not exist!
                    pass
                else:
                    for details in curr_stackvar.offset_and_details.values():
                        if details.def_at is None or not self._check_store_precedes_load(
                                details.def_at, self.expr_defat):
                            self.out_dated = True
                            break
                    if not self.out_dated:
                        # if there has been a stack store whose address cannot be resolved or concretized, we see if
                        # this store happens after the current definition. if so, we mark it as out-dated
                        if self.state.last_stack_store is not None \
                                and not self._check_store_precedes_load(CodeLocation(*self.state.last_stack_store[:2]),
                                                                        self.expr_defat):
                            self.out_dated = True
            else:
                # in cases where expr.addr cannot be resolved to a concrete stack offset, we play safe and assume
                # it's outdated
                self.out_dated = True
        elif isinstance(expr.addr, Expr.Const) \
                and self.state.global_stores \
                and any(self._check_global_store_conflicts_load(store_block_addr, store_stmt_idx, addr, store,
                                                            self.expr_defat.block_addr, self.expr_defat.stmt_idx, expr)
                        for store_block_addr, store_stmt_idx, addr, store in self.state.global_stores):
            self.out_dated = True
        elif not isinstance(expr.addr, (Expr.StackBaseOffset, Expr.Const)) \
                and (
                self.state.global_stores and
                not all(self._check_store_precedes_load(CodeLocation(store_block_addr, store_stmt_idx), self.expr_defat)
                        for store_block_addr, store_stmt_idx, addr, store in self.state.global_stores) or
                self.state.last_stack_store is not None and
                not self._check_store_precedes_load(CodeLocation(*self.state.last_stack_store[:2]),
                                                    self.expr_defat)
        ):
            # check both stack and global stores if the load address is unknown
            self.out_dated = True
        else:
            super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_VEXCCallExpression(self, expr_idx: int, expr: Expr.VEXCCallExpression, stmt_idx: int,
                                   stmt: Stmt.Statement, block: Optional[Block]):
        if self.avoid is not None:
            if any(op == self.avoid for op in expr.operands):
                self.out_dated = True

        if not self.out_dated:
            super()._handle_VEXCCallExpression(expr_idx, expr, stmt_idx, stmt, block)

    @staticmethod
    def _reg_overlap(reg0: Expr.Register, reg1: Expr.Register) -> bool:
        if reg0.reg_offset <= reg1.reg_offset < reg0.reg_offset + reg0.size:
            return True
        if reg1.reg_offset <= reg0.reg_offset < reg1.reg_offset + reg1.size:
            return True
        return False



import multiprocessing
import platform


def mp_context():
    system = platform.system()
    spawn_methods = {
        "Windows": "spawn",
        "Linux": "fork",
        # Python<3.8 defaults to fork
        # https://bugs.python.org/issue33725
        "Darwin": "spawn",
    }
    spawn_method = spawn_methods.get(system, "fork")  # default to fork on other platforms
    return multiprocessing.get_context(spawn_method)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("qwave.dll")
prototypes = \
    {
        # 
        'QOSCreateHandle': SimTypeFunction([SimTypePointer(SimStruct({"MajorVersion": SimTypeShort(signed=False, label="UInt16"), "MinorVersion": SimTypeShort(signed=False, label="UInt16")}, name="QOS_VERSION", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Version", "QOSHandle"]),
        # 
        'QOSCloseHandle': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle"]),
        # 
        'QOSStartTrackingClient': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"sa_family": SimTypeShort(signed=False, label="UInt16"), "sa_data": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 14)}, name="SOCKADDR", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "DestAddr", "Flags"]),
        # 
        'QOSStopTrackingClient': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"sa_family": SimTypeShort(signed=False, label="UInt16"), "sa_data": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 14)}, name="SOCKADDR", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "DestAddr", "Flags"]),
        # 
        'QOSEnumerateFlows': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Size", "Buffer"]),
        # 
        'QOSAddSocketToFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimStruct({"sa_family": SimTypeShort(signed=False, label="UInt16"), "sa_data": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 14)}, name="SOCKADDR", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="QOS_TRAFFIC_TYPE"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Socket", "DestAddr", "TrafficType", "Flags", "FlowId"]),
        # 
        'QOSRemoveSocketFromFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Socket", "FlowId", "Flags"]),
        # 
        'QOSSetFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="QOS_SET_FLOW"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "FlowId", "Operation", "Size", "Buffer", "Flags", "Overlapped"]),
        # 
        'QOSQueryFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="QOS_QUERY_FLOW"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "FlowId", "Operation", "Size", "Buffer", "Flags", "Overlapped"]),
        # 
        'QOSNotifyFlow': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="QOS_NOTIFY_FLOW"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "FlowId", "Operation", "Size", "Buffer", "Flags", "Overlapped"]),
        # 
        'QOSCancel': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["QOSHandle", "Overlapped"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("netsh.dll")
prototypes = \
    {
        # 
        'MatchEnumTag': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"pwszToken": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwValue": SimTypeInt(signed=False, label="UInt32")}, name="TOKEN_VALUE", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "pwcArg", "dwNumArg", "pEnumTable", "pdwValue"]),
        # 
        'MatchToken': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwszUserToken", "pwszCmdToken"]),
        # 
        'PreprocessCommand': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"pwszTag": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwRequired": SimTypeInt(signed=False, label="UInt32"), "bPresent": SimTypeInt(signed=True, label="Int32")}, name="TAG_TYPE", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "ppwcArguments", "dwCurrentIndex", "dwArgCount", "pttTags", "dwTagCount", "dwMinArgs", "dwMaxArgs", "pdwTagType"]),
        # 
        'PrintError': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "dwErrId"]),
        # 
        'PrintMessageFromModule': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hModule", "dwMsgId"]),
        # 
        'PrintMessage': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pwszFormat"]),
        # 
        'RegisterContext': SimTypeFunction([SimTypePointer(SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "_ullAlign": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "pwszContext": SimTypePointer(SimTypeChar(label="Char"), offset=0), "guidHelper": SimTypeBottom(label="Guid"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "ulPriority": SimTypeInt(signed=False, label="UInt32"), "ulNumTopCmds": SimTypeInt(signed=False, label="UInt32"), "pTopCmds": SimTypePointer(SimStruct({"pwszCmdToken": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pfnCmdHandler": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pwszMachine", "ppwcArguments", "dwCurrentIndex", "dwArgCount", "dwFlags", "pvData", "pbDone"]), offset=0), "dwShortCmdHelpToken": SimTypeInt(signed=False, label="UInt32"), "dwCmdHlpToken": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pOsVersionCheck": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CIMOSType", "CIMOSProductSuite", "CIMOSVersion", "CIMOSBuildNumber", "CIMServicePackMajorVersion", "CIMServicePackMinorVersion", "uiReserved", "dwReserved"]), offset=0)}, name="CMD_ENTRY", pack=False, align=None), offset=0), "ulNumGroups": SimTypeInt(signed=False, label="UInt32"), "pCmdGroups": SimTypePointer(SimStruct({"pwszCmdGroupToken": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwShortCmdHelpToken": SimTypeInt(signed=False, label="UInt32"), "ulCmdGroupSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pCmdGroup": SimTypePointer(SimStruct({"pwszCmdToken": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pfnCmdHandler": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pwszMachine", "ppwcArguments", "dwCurrentIndex", "dwArgCount", "dwFlags", "pvData", "pbDone"]), offset=0), "dwShortCmdHelpToken": SimTypeInt(signed=False, label="UInt32"), "dwCmdHlpToken": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pOsVersionCheck": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CIMOSType", "CIMOSProductSuite", "CIMOSVersion", "CIMOSBuildNumber", "CIMServicePackMajorVersion", "CIMServicePackMinorVersion", "uiReserved", "dwReserved"]), offset=0)}, name="CMD_ENTRY", pack=False, align=None), offset=0), "pOsVersionCheck": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CIMOSType", "CIMOSProductSuite", "CIMOSVersion", "CIMOSBuildNumber", "CIMServicePackMajorVersion", "CIMServicePackMinorVersion", "uiReserved", "dwReserved"]), offset=0)}, name="CMD_GROUP_ENTRY", pack=False, align=None), offset=0), "pfnCommitFn": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwAction"]), offset=0), "pfnDumpFn": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pwszRouter", "ppwcArguments", "dwArgCount", "pvData"]), offset=0), "pfnConnectFn": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pwszMachine"]), offset=0), "pReserved": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pfnOsVersionCheck": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["CIMOSType", "CIMOSProductSuite", "CIMOSVersion", "CIMOSBuildNumber", "CIMServicePackMajorVersion", "CIMServicePackMinorVersion", "uiReserved", "dwReserved"]), offset=0)}, name="NS_CONTEXT_ATTRIBUTES", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pChildContext"]),
        # 
        'RegisterHelper': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwReserved": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "_ullAlign": SimTypeLongLong(signed=False, label="UInt64")}, name="<anon>", label="None"), "guidHelper": SimTypeBottom(label="Guid"), "pfnStart": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pguidParent", "dwVersion"]), offset=0), "pfnStop": SimTypePointer(SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwReserved"]), offset=0)}, name="NS_HELPER_ATTRIBUTES", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pguidParentContext", "pfnRegisterSubContext"]),
    }

lib.set_prototypes(prototypes)



import platform
from typing import Optional, Union, Dict, List, Generic, TypeVar


# Ref: https://github.com/angr/angr/pull/3471#issuecomment-1236515950
if platform.python_implementation() == "PyPy":
    LIST2DICT_THRESHOLD = 96
else:
    # cpython
    LIST2DICT_THRESHOLD = 2048


VT = TypeVar("VT")


class DynamicDictList(Generic[VT]):
    """
    A list-like container class that internally uses dicts to store values when the number of values is less than the
    threshold `LIST2DICT_THRESHOLD`. Keys must be ints.

    The default thresholds are determined according to experiments described at
    https://github.com/angr/angr/pull/3471#issuecomment-1236515950.
    """

    __slots__ = ('list_content', 'dict_content', 'max_size')

    def __init__(self,
                 max_size: Optional[int]=None,
                 content: Optional[Union['DynamicDictList',Dict[int,VT],List[VT]]]=None):
        self.list_content: Optional[List[VT]] = None
        self.dict_content: Optional[Dict[int,VT]] = None
        self.max_size = max_size

        if content:
            self._initialize_content(content)
        else:
            self.dict_content = { }

    def _initialize_content(self, content) -> None:

        if isinstance(content, DynamicDictList):
            # make a copy
            self.list_content = list(content.list_content) if content.list_content is not None else None
            self.dict_content = dict(content.dict_content) if content.dict_content is not None else None
            return

        # initializing from a list or a dict
        if len(content) < LIST2DICT_THRESHOLD:
            # use a dict
            if isinstance(content, list):
                self.dict_content = dict(enumerate(content))
            else:
                self.dict_content = dict(content)
        else:
            # use a list
            if isinstance(content, list):
                self.list_content = list(content)
            else:
                self.list_content = [None] * self.max_size
                for k, v in content.items():
                    self.list_content[k] = v

    def real_length(self) -> int:
        if self.dict_content is not None:
            return len(self.dict_content)
        return len(self.list_content)

    def __len__(self) -> int:
        return self.max_size

    def __getitem__(self, key: int) -> VT:
        if self.dict_content is not None:
            return self.dict_content.get(key, None)
        return self.list_content[key]

    def __setitem__(self, key: int, value: VT) -> None:
        if self.dict_content is not None:
            self.dict_content[key] = value
            if len(self.dict_content) >= LIST2DICT_THRESHOLD:
                # switch
                self._initialize_content(self.dict_content)
                self.dict_content = None
        else:
            self.list_content[key] = value

    def __iter__(self):
        if self.dict_content is not None:
            for i in range(self.max_size):
                yield self.dict_content.get(i, None)
        else:
            yield from self.list_content



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("icmui.dll")
prototypes = \
    {
        # 
        'SetupColorMatchingW': SimTypeFunction([SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "pSourceName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pDisplayName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pPrinterName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwRenderIntent": SimTypeInt(signed=False, label="UInt32"), "dwProofingIntent": SimTypeInt(signed=False, label="UInt32"), "pMonitorProfile": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ccMonitorProfile": SimTypeInt(signed=False, label="UInt32"), "pPrinterProfile": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ccPrinterProfile": SimTypeInt(signed=False, label="UInt32"), "pTargetProfile": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ccTargetProfile": SimTypeInt(signed=False, label="UInt32"), "lpfnHook": SimTypeBottom(label="DLGPROC"), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpfnApplyCallback": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="COLORMATCHSETUPW"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]), offset=0), "lParamApplyCallback": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="COLORMATCHSETUPW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pcms"]),
        # 
        'SetupColorMatchingA': SimTypeFunction([SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "hwndOwner": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "pSourceName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pDisplayName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pPrinterName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "dwRenderIntent": SimTypeInt(signed=False, label="UInt32"), "dwProofingIntent": SimTypeInt(signed=False, label="UInt32"), "pMonitorProfile": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ccMonitorProfile": SimTypeInt(signed=False, label="UInt32"), "pPrinterProfile": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ccPrinterProfile": SimTypeInt(signed=False, label="UInt32"), "pTargetProfile": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ccTargetProfile": SimTypeInt(signed=False, label="UInt32"), "lpfnHook": SimTypeBottom(label="DLGPROC"), "lParam": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), "lpfnApplyCallback": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="COLORMATCHSETUPA"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]), offset=0), "lParamApplyCallback": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="COLORMATCHSETUPA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pcms"]),
    }

lib.set_prototypes(prototypes)




import logging
from typing import Optional

from archinfo.arch_soot import (ArchSoot, SootAddressDescriptor, SootArgument,
                                SootMethodDescriptor)

from . import JNISimProcedure
from ...calling_conventions import SimCCSoot
from ...engines.soot.method_dispatcher import resolve_method

l = logging.getLogger('angr.procedures.java_jni.callmethod')

# pylint: disable=arguments-differ,unused-argument

#
# GetMethodID / GetStaticMethodID
#

class GetMethodID(JNISimProcedure):

    return_ty = 'reference'

    def run(self, ptr_env, class_, ptr_method_name, ptr_method_sig):

        method_class = self.state.jni_references.lookup(class_)
        method_name = self._load_string_from_native_memory(ptr_method_name)
        method_sig = self._load_string_from_native_memory(ptr_method_sig)

        # derive parameter type from signature
        params, _ = ArchSoot.decode_method_signature(method_sig)

        # create method ID and return reference to it
        method_id = SootMethodDescriptor(method_class.name, method_name, params)
        return self.state.jni_references.create_new_reference(method_id)

#
# Call<Type>Method / CallNonvirtual<Type>Method / CallStatic<Type>Method
#

class CallMethodBase(JNISimProcedure):

    return_ty: Optional[str] = None

    def _invoke(self, method_id, obj=None, dynamic_dispatch=True, args_in_array=None):
        # get invoke target
        class_name = obj.type if dynamic_dispatch else method_id.class_name
        invoke_target = resolve_method(self.state, method_id.name, class_name, method_id.params)
        invoke_addr = SootAddressDescriptor(invoke_target, 0, 0)

        # get args
        no_of_args = len(invoke_target.params)
        if args_in_array is not None:
            arg_values = self._get_arg_values_from_array(args_in_array, no_of_args)
        else:
            arg_values = self._get_arg_values(no_of_args)

        # setup java args
        java_args = self._setup_java_args(arg_values, invoke_target, this_ref=obj)

        # call java method
        # => after returning, the execution will be continued in _return_result_of_computation
        self.call(invoke_addr, java_args, "return_from_invocation", cc=SimCCSoot(ArchSoot()))

    def _get_arg_values(self, no_of_args):
        return [self.va_arg('void*') for _ in range(no_of_args)]

    def _get_arg_values_from_array(self, array, no_of_args):
        return self._load_from_native_memory(addr=array, data_size=self.arch.bytes,
                                             no_of_elements=no_of_args, return_as_list=True)

    def _setup_java_args(self, arg_values, method_id, this_ref=None):
        args = []

        # if available, add 'this' reference
        if this_ref:
            args += [ SootArgument(this_ref, this_ref.type, is_this_ref=True) ]

        # function arguments
        for arg_value_, arg_type in zip(arg_values, method_id.params):

            if arg_type in ArchSoot.primitive_types:
                # argument has a primitive integral type
                # => cast native value to java type
                arg_value = self.project.simos.cast_primitive(self.state,
                                                              value=arg_value_,
                                                              to_type=arg_type)

            else:
                # argument has a relative type
                # => lookup java object
                arg_value = self.state.jni_references.lookup(arg_value_)

            args += [ SootArgument(arg_value, arg_type) ]

        return args

    def _return_from_invocation(self):
        if self.return_ty != 'void':
            ret_value = self.state.javavm_registers.load('invoke_return_value')
            if self.return_ty == 'reference':
                return self.state.jni_references.create_new_reference(ret_value)
            else:
                return ret_value

#
# Call<Type>Method
#

class CallMethod(CallMethodBase):
    def run(self, ptr_env, obj_, method_id_):
        method_id = self.state.jni_references.lookup(method_id_)
        obj = self.state.jni_references.lookup(obj_)
        self._invoke(method_id, obj, dynamic_dispatch=True)

    def return_from_invocation(self, ptr_env, obj_, method_id_):
        return self._return_from_invocation()

class CallMethodA(CallMethodBase):
    def run(self, ptr_env, obj_, method_id_, ptr_args):
        method_id = self.state.jni_references.lookup(method_id_)
        obj = self.state.jni_references.lookup(obj_)
        self._invoke(method_id, obj, dynamic_dispatch=True, args_in_array=ptr_args)

    def return_from_invocation(self, ptr_env, obj_, method_id_, ptr_args):
        return self._return_from_invocation()

class CallObjectMethod(CallMethod):
    return_ty = 'reference'
class CallBooleanMethod(CallMethod):
    return_ty = 'boolean'
class CallByteMethod(CallMethod):
    return_ty = 'byte'
class CallCharMethod(CallMethod):
    return_ty = 'char'
class CallShortMethod(CallMethod):
    return_ty = 'short'
class CallIntMethod(CallMethod):
    return_ty = 'int'
class CallLongMethod(CallMethod):
    return_ty = 'long'
class CallVoidMethod(CallMethod):
    return_ty = 'void'
class CallObjectMethodA(CallMethodA):
    return_ty = 'reference'
class CallBooleanMethodA(CallMethodA):
    return_ty = 'boolean'
class CallByteMethodA(CallMethodA):
    return_ty = 'byte'
class CallCharMethodA(CallMethodA):
    return_ty = 'char'
class CallShortMethodA(CallMethodA):
    return_ty = 'short'
class CallIntMethodA(CallMethodA):
    return_ty = 'int'
class CallLongMethodA(CallMethodA):
    return_ty = 'long'
class CallVoidMethodA(CallMethodA):
    return_ty = 'void'

#
# CallNonVirtual<Type>Method
#

class CallNonvirtualMethod(CallMethodBase):
    def run(self, ptr_env, obj_, class_, method_id_):
        method_id = self.state.jni_references.lookup(method_id_)
        obj = self.state.jni_references.lookup(obj_)
        self._invoke(method_id, obj, dynamic_dispatch=False)

    def return_from_invocation(self, ptr_env, obj_, class_, method_id_):
        return self._return_from_invocation()

class CallNonvirtualMethodA(CallMethodBase):
    def run(self, ptr_env, obj_, method_id_, ptr_args):
        method_id = self.state.jni_references.lookup(method_id_)
        obj = self.state.jni_references.lookup(obj_)
        self._invoke(method_id, obj, dynamic_dispatch=False, args_in_array=ptr_args)

    def return_from_invocation(self, ptr_env, obj_, method_id_, ptr_args):
        return self._return_from_invocation()

class CallNonvirtualObjectMethod(CallNonvirtualMethod):
    return_ty = 'reference'
class CallNonvirtualBooleanMethod(CallNonvirtualMethod):
    return_ty = 'boolean'
class CallNonvirtualByteMethod(CallNonvirtualMethod):
    return_ty = 'byte'
class CallNonvirtualCharMethod(CallNonvirtualMethod):
    return_ty = 'char'
class CallNonvirtualShortMethod(CallNonvirtualMethod):
    return_ty = 'short'
class CallNonvirtualIntMethod(CallNonvirtualMethod):
    return_ty = 'int'
class CallNonvirtualLongMethod(CallNonvirtualMethod):
    return_ty = 'long'
class CallNonvirtualVoidMethod(CallNonvirtualMethod):
    return_ty = 'void'
class CallNonvirtualObjectMethodA(CallNonvirtualMethodA):
    return_ty = 'reference'
class CallNonvirtualBooleanMethodA(CallNonvirtualMethodA):
    return_ty = 'boolean'
class CallNonvirtualByteMethodA(CallNonvirtualMethodA):
    return_ty = 'byte'
class CallNonvirtualCharMethodA(CallNonvirtualMethodA):
    return_ty = 'char'
class CallNonvirtualShortMethodA(CallNonvirtualMethodA):
    return_ty = 'short'
class CallNonvirtualIntMethodA(CallNonvirtualMethodA):
    return_ty = 'int'
class CallNonvirtualLongMethodA(CallNonvirtualMethodA):
    return_ty = 'long'
class CallNonvirtualVoidMethodA(CallNonvirtualMethodA):
    return_ty = 'void'

#
# CallStatic<Type>Method
#

class CallStaticMethod(CallMethodBase):
    def run(self, ptr_env, class_, method_id_):
        method_id = self.state.jni_references.lookup(method_id_)
        self._invoke(method_id, dynamic_dispatch=False)

    def return_from_invocation(self, ptr_env, class_, method_id_):
        return self._return_from_invocation()

class CallStaticMethodA(CallMethodBase):
    def run(self, ptr_env, obj_, method_id_, ptr_args):
        method_id = self.state.jni_references.lookup(method_id_)
        self._invoke(method_id, dynamic_dispatch=False, args_in_array=ptr_args)

    def return_from_invocation(self, ptr_env, class_, method_id_, ptr_args):
        return self._return_from_invocation()

class CallStaticObjectMethod(CallStaticMethod):
    return_ty = 'reference'
class CallStaticBooleanMethod(CallStaticMethod):
    return_ty = 'boolean'
class CallStaticByteMethod(CallStaticMethod):
    return_ty = 'byte'
class CallStaticCharMethod(CallStaticMethod):
    return_ty = 'char'
class CallStaticShortMethod(CallStaticMethod):
    return_ty = 'short'
class CallStaticIntMethod(CallStaticMethod):
    return_ty = 'int'
class CallStaticLongMethod(CallStaticMethod):
    return_ty = 'long'
class CallStaticVoidMethod(CallStaticMethod):
    return_ty = 'void'
class CallStaticObjectMethodA(CallStaticMethodA):
    return_ty = 'reference'
class CallStaticBooleanMethodA(CallStaticMethodA):
    return_ty = 'boolean'
class CallStaticByteMethodA(CallStaticMethodA):
    return_ty = 'byte'
class CallStaticCharMethodA(CallStaticMethodA):
    return_ty = 'char'
class CallStaticShortMethodA(CallStaticMethodA):
    return_ty = 'short'
class CallStaticIntMethodA(CallStaticMethodA):
    return_ty = 'int'
class CallStaticLongMethodA(CallStaticMethodA):
    return_ty = 'long'
class CallStaticVoidMethodA(CallStaticMethodA):
    return_ty = 'void'



import logging

from ailment import Expr

from .engine_base import SimplifierAILEngine, SimplifierAILState
from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)

class MultiSimplifierAILEngine(SimplifierAILEngine):
    """
    An AIL pass for the multi simplifier
    """

    def _ail_handle_Add(self, expr):

        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])


        # x + x = 2*x
        if type(operand_0) in [Expr.Convert, Expr.Register]:
            if isinstance(operand_1, (Expr.Convert, Expr.Register)):
                if operand_0 == operand_1:
                    count = Expr.Const(expr.idx, None, 2, 8)
                    return Expr.BinaryOp(expr.idx, 'Mul', [operand_1, count], expr.signed, **expr.tags)
        # 2*x + x = 3*x
        if Expr.BinaryOp in [type(operand_0), type(operand_1)]:
            if isinstance(operand_1, Expr.BinaryOp) and operand_1.op == 'Mul' and \
                    (not isinstance(operand_0, Expr.BinaryOp) or
                        (isinstance(operand_0, Expr.BinaryOp) and operand_0.op != 'Mul')):
                x0 = operand_0
                x1_index = 0 if isinstance(operand_1.operands[1], Expr.Const) else 1
                x1 = operand_1.operands[x1_index]
                const_x1 = operand_1.operands[1-x1_index]
                if x0 == x1:
                    new_const = Expr.Const(const_x1.idx, None, const_x1.value+1, const_x1.bits)
                    new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x0, new_const], expr.signed, **expr.tags)
                    return new_expr
            elif isinstance(operand_0, Expr.BinaryOp) and operand_0.op == 'Mul' and \
                    (not isinstance(operand_1, Expr.BinaryOp) or
                        (isinstance(operand_1, Expr.BinaryOp) and operand_1.op != 'Mul')):

                x1 = operand_1
                x0_index = 0 if isinstance(operand_0.operands[1], Expr.Const) else 1
                x0 = operand_0.operands[x0_index]
                const_x0 = operand_0.operands[1-x0_index]
                if x0 == x1:
                    new_const = Expr.Const(const_x0.idx, None, const_x0.value+1, const_x0.bits)
                    new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x1, new_const], expr.signed, **expr.tags)
                    return new_expr
            # 2*x + 3*x = 5*x
            elif isinstance(operand_0, Expr.BinaryOp) and isinstance(operand_1, Expr.BinaryOp) and \
                    operand_0.op == 'Mul' and operand_1.op == 'Mul':
                if Expr.Const in [type(operand_0.operands[0]), type(operand_0.operands[1])] \
                    and Expr.Const in [type(operand_1.operands[0]), type(operand_1.operands[1])]:
                    x0_index = 0 if isinstance(operand_0.operands[1], Expr.Const) else 1
                    x0 = operand_0.operands[x0_index]
                    const_x0 = operand_0.operands[1-x0_index]

                    x1_index = 0 if isinstance(operand_1.operands[1], Expr.Const) else 1
                    x1 = operand_1.operands[x1_index]
                    const_x1 = operand_1.operands[1-x1_index]
                    if x0 == x1:
                        new_const = Expr.Const(const_x1.idx, None, const_x1.value+const_x0.value, const_x1.bits)
                        new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x0, new_const], expr.signed, **expr.tags)
                        return new_expr

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, 'Add', [operand_0, operand_1], expr.signed, **expr.tags)
        return expr


    def _ail_handle_Sub(self, expr):

        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        # x + x = 2*x
        if type(operand_0) in [Expr.Convert, Expr.Register]:
            if isinstance(operand_1, (Expr.Convert, Expr.Register)):
                if operand_0 == operand_1:
                    count = Expr.Const(expr.idx, None, 0, 8)
                    new_expr = Expr.BinaryOp(expr.idx, 'Mul', [operand_1, count], expr.signed, **expr.tags)
                    return new_expr

        # 2*x - x = x
        if Expr.BinaryOp in [type(operand_0), type(operand_1)]:
            if isinstance(operand_1, Expr.BinaryOp) and operand_1.op == 'Mul' and \
                (not isinstance(operand_0, Expr.BinaryOp) or \
                    (isinstance(operand_0, Expr.BinaryOp) and operand_0.op != 'Mul')):

                    x0 = operand_0
                    x1_index = 0 if isinstance(operand_1.operands[1], Expr.Const) else 1
                    x1 = operand_1.operands[x1_index]
                    const_x1 = operand_1.operands[1-x1_index]
                    if x0 == x1:
                        new_const = Expr.Const(const_x1.idx, None, const_x1.value-1, const_x1.bits)
                        new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x0, new_const], expr.signed, **expr.tags)
                        return new_expr
            elif isinstance(operand_0, Expr.BinaryOp) and operand_0.op == 'Mul' and \
                (not isinstance(operand_1, Expr.BinaryOp) or \
                    (isinstance(operand_1, Expr.BinaryOp) and operand_1.op != 'Mul')):

                    x1 = operand_1
                    x0_index = 0 if isinstance(operand_0.operands[1], Expr.Const) else 1
                    x0 = operand_0.operands[x0_index]
                    const_x0 = operand_0.operands[1-x0_index]
                    if x0 == x1:
                        new_const = Expr.Const(const_x0.idx, None, const_x0.value-1, const_x0.bits)
                        new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x1, new_const], expr.signed, **expr.tags)
                        return new_expr
            # 3*x - 2*x = x
            elif isinstance(operand_0, Expr.BinaryOp) and isinstance(operand_1, Expr.BinaryOp) and \
                operand_0.op == 'Mul' and operand_1.op == 'Mul':
                if Expr.Const in [type(operand_0.operands[0]), type(operand_0.operands[1])] \
                    and Expr.Const in [type(operand_1.operands[0]), type(operand_1.operands[1])]:
                    x0_index = 0 if isinstance(operand_0.operands[1], Expr.Const) else 1
                    x0 = operand_0.operands[x0_index]
                    const_x0 = operand_0.operands[1-x0_index]

                    x1_index = 0 if isinstance(operand_1.operands[1], Expr.Const) else 1
                    x1 = operand_1.operands[x1_index]
                    const_x1 = operand_1.operands[1-x1_index]
                    if x0 == x1:
                        new_const = Expr.Const(const_x1.idx, None, const_x0.value-const_x1.value, const_x1.bits)
                        new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x0, new_const], expr.signed, **expr.tags)
                        return new_expr

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, 'Sub', [operand_0, operand_1], expr.signed, **expr.tags)
        return expr

    def _ail_handle_Shl(self, expr):

        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        if isinstance(operand_1, Expr.Const):
            new_operand = Expr.Const(operand_1.idx, None, 2**operand_1.value, operand_0.bits)
            return Expr.BinaryOp(expr.idx, 'Mul', [operand_0, new_operand], expr.signed, **expr.tags)

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, 'Shl', [operand_0, operand_1], expr.signed, **expr.tags)
        return expr

    def _ail_handle_Mul(self, expr):

        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        if Expr.Const in [type(operand_0), type(operand_1)]:
            if Expr.BinaryOp in [type(operand_0), type(operand_1)]:
                const_, x0 = (operand_0, operand_1) if isinstance(operand_0, Expr.Const) else (operand_1, operand_0)
                if x0.op == 'Mul' and Expr.Const in [type(x0.operands[0]), type(x0.operands[1])]:
                    if isinstance(x0.operands[0], Expr.Const):
                        const_x0, x = x0.operands[0], x0.operands[1]
                    else:
                        const_x0, x = x0.operands[1], x0.operands[0]
                    new_const = Expr.Const(const_.idx, None, const_.value*const_x0.value, const_.bits)
                    new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x, new_const], expr.signed, **expr.tags)
                    return new_expr
            elif isinstance(operand_0, Expr.Convert) \
                and isinstance(operand_0.operand, Expr.BinaryOp) \
                    and operand_0.operand.op == 'Mul' \
                    and isinstance(operand_0.operand.operands[1], Expr.Const):
                x = operand_0.operand.operands[0]
                new_const = Expr.Const(operand_1.idx, None,
                                    operand_1.value*operand_0.operand.operands[1].value, operand_1.bits)
                new_expr = Expr.BinaryOp(expr.idx, 'Mul', [x, new_const], expr.signed, **expr.tags)
                return new_expr

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, 'Mul', [operand_0, operand_1], expr.signed, **expr.tags)
        return expr

class MultiSimplifier(OptimizationPass):
    """
    Implements several different arithmetic optimizations.
    """

    ARCHES = ["X86", "AMD64"]
    PLATFORMS = ["linux", "windows"]
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Simplify various arithmetic expressions"
    DESCRIPTION = __doc__.strip()

    def __init__(self, func, **kwargs):

        super().__init__(func, **kwargs)

        self.state = SimplifierAILState(self.project.arch)
        self.engine = MultiSimplifierAILEngine()

        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):

        for block in list(self._graph.nodes()):
            new_block = block
            old_block = None

            while new_block != old_block:
                old_block = new_block
                new_block = self.engine.process(state=self.state.copy(), block=old_block.copy())
                _l.debug("new block: %s", new_block.statements)

            self._update_block(block, new_block)



from typing import Set, Optional, Union, TYPE_CHECKING

from ...knowledge_plugins.key_definitions import LiveDefinitions
from .. import register_analysis
from .reaching_definitions import ReachingDefinitionsAnalysis


if TYPE_CHECKING:
    from angr.knowledge_plugins.key_definitions.definition import Definition
    from angr.storage.memory_object import SimMemoryObject
    from angr.storage.memory_mixins import MultiValuedMemory
    from angr.storage.memory_mixins.paged_memory.pages import MVListPage


def get_all_definitions(region: 'MultiValuedMemory') -> Set['Definition']:

    all_defs: Set['Definition'] = set()

    # MultiValuedMemory only uses ListPage internally
    for page in region._pages.values():
        page: 'MVListPage'

        for idx in page.stored_offset:
            cnt_set: Optional[Union['SimMemoryObject', Set['SimMemoryObject']]] = page.content[idx]
            if cnt_set is None:
                continue
            elif not type(cnt_set) is set:
                cnt_set = { cnt_set }
            for cnt in cnt_set:
                for def_ in LiveDefinitions.extract_defs(cnt.object):
                    all_defs.add(def_)

    return all_defs


register_analysis(ReachingDefinitionsAnalysis, 'ReachingDefinitions')
